"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.sequenceReadonlyArray = exports.traverseReadonlyArray = exports.traverseReadonlyNonEmptyArray = exports.traverseReadonlyArrayWithIndex = exports.traverseReadonlyNonEmptyArrayWithIndex = exports.apT = exports.tupled = exports.ApT = exports.apS = exports.bind = exports.let = exports.bindTo = exports.Do = exports.ChainRec = exports.FromIO = exports.chainFirst = exports.Monad = exports.Chain = exports.Applicative = exports.apSecond = exports.apFirst = exports.Apply = exports.Pointed = exports.flap = exports.Functor = exports.flatten = exports.chainRec = exports.chain = exports.of = exports.ap = exports.map = void 0;
/**
 * ```ts
 * interface IO<A> {
 *   (): A
 * }
 * ```
 *
 * `IO<A>` represents a non-deterministic synchronous computation that can cause side effects, yields a value of
 * type `A` and **never fails**.
 *
 * If you want to represent a synchronous computation that may fail, please see `IOEither`.
 * If you want to represent a synchronous computation that may yield nothing, please see `IOOption`.
 *
 * @since 3.0.0
 */
var ApplyModule = __importStar(require("./Apply"));
var ChainModule = __importStar(require("./Chain"));
var function_1 = require("./function");
var FunctorModule = __importStar(require("./Functor"));
var _ = __importStar(require("./internal"));
// -------------------------------------------------------------------------------------
// type class members
// -------------------------------------------------------------------------------------
/**
 * `map` can be used to turn functions `(a: A) => B` into functions `(fa: F<A>) => F<B>` whose argument and return types
 * use the type constructor `F` to represent some computational context.
 *
 * @category Functor
 * @since 3.0.0
 */
var map = function (f) { return function (fa) { return function () { return f(fa()); }; }; };
exports.map = map;
/**
 * Apply a function to an argument under a type constructor.
 *
 * @category Apply
 * @since 3.0.0
 */
var ap = function (fa) { return function (fab) { return function () { return fab()(fa()); }; }; };
exports.ap = ap;
/**
 * @category Pointed
 * @since 3.0.0
 */
exports.of = function_1.constant;
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation.
 *
 * @category Chain
 * @since 3.0.0
 */
var chain = function (f) { return function (ma) { return function () { return f(ma())(); }; }; };
exports.chain = chain;
/**
 * @category ChainRec
 * @since 3.0.0
 */
var chainRec = function (f) { return function (a) { return function () {
    var e = f(a)();
    while (_.isLeft(e)) {
        e = f(e.left)();
    }
    return e.right;
}; }; };
exports.chainRec = chainRec;
/**
 * Derivable from `Chain`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
exports.flatten = (0, exports.chain)(function_1.identity);
// -------------------------------------------------------------------------------------
// instances
// -------------------------------------------------------------------------------------
/**
 * @category instances
 * @since 3.0.0
 */
exports.Functor = {
    map: exports.map
};
/**
 * Derivable from `Functor`.
 *
 * @category combinators
 * @since 3.0.0
 */
exports.flap = FunctorModule.flap(exports.Functor);
/**
 * @category instances
 * @since 3.0.0
 */
exports.Pointed = {
    of: exports.of
};
/**
 * @category instances
 * @since 3.0.0
 */
exports.Apply = {
    map: exports.map,
    ap: exports.ap
};
/**
 * Combine two effectful actions, keeping only the result of the first.
 *
 * Derivable from `Apply`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
exports.apFirst = ApplyModule.apFirst(exports.Apply);
/**
 * Combine two effectful actions, keeping only the result of the second.
 *
 * Derivable from `Apply`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
exports.apSecond = ApplyModule.apSecond(exports.Apply);
/**
 * @category instances
 * @since 3.0.0
 */
exports.Applicative = {
    map: exports.map,
    ap: exports.ap,
    of: exports.of
};
/**
 * @category instances
 * @since 3.0.0
 */
exports.Chain = {
    map: exports.map,
    chain: exports.chain
};
/**
 * @category instances
 * @since 3.0.0
 */
exports.Monad = {
    map: exports.map,
    of: exports.of,
    chain: exports.chain
};
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation and
 * keeping only the result of the first.
 *
 * Derivable from `Chain`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
exports.chainFirst = 
/*#__PURE__*/ ChainModule.chainFirst(exports.Chain);
/**
 * @category instances
 * @since 3.0.0
 */
exports.FromIO = {
    fromIO: function_1.identity
};
/**
 * @category instances
 * @since 3.0.0
 */
exports.ChainRec = {
    chainRec: exports.chainRec
};
// -------------------------------------------------------------------------------------
// do notation
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
exports.Do = (0, exports.of)(_.emptyRecord);
/**
 * @since 3.0.0
 */
exports.bindTo = 
/*#__PURE__*/ FunctorModule.bindTo(exports.Functor);
var let_ = 
/*#__PURE__*/ FunctorModule.let(exports.Functor);
exports.let = let_;
/**
 * @since 3.0.0
 */
exports.bind = 
/*#__PURE__*/ ChainModule.bind(exports.Chain);
// -------------------------------------------------------------------------------------
// sequence S
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
exports.apS = 
/*#__PURE__*/ ApplyModule.apS(exports.Apply);
// -------------------------------------------------------------------------------------
// sequence T
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
exports.ApT = (0, exports.of)(_.emptyReadonlyArray);
/**
 * @since 3.0.0
 */
exports.tupled = FunctorModule.tupled(exports.Functor);
/**
 * @since 3.0.0
 */
exports.apT = 
/*#__PURE__*/ ApplyModule.apT(exports.Apply);
// -------------------------------------------------------------------------------------
// array utils
// -------------------------------------------------------------------------------------
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverseWithIndex(Apply)`.
 *
 * @since 3.0.0
 */
var traverseReadonlyNonEmptyArrayWithIndex = function (f) {
    return function (as) {
        return function () {
            var out = [f(0, _.head(as))()];
            for (var i = 1; i < as.length; i++) {
                out.push(f(i, as[i])());
            }
            return out;
        };
    };
};
exports.traverseReadonlyNonEmptyArrayWithIndex = traverseReadonlyNonEmptyArrayWithIndex;
/**
 * Equivalent to `ReadonlyArray#traverseWithIndex(Applicative)`.
 *
 * @since 3.0.0
 */
var traverseReadonlyArrayWithIndex = function (f) {
    var g = (0, exports.traverseReadonlyNonEmptyArrayWithIndex)(f);
    return function (as) { return (_.isNonEmpty(as) ? g(as) : exports.ApT); };
};
exports.traverseReadonlyArrayWithIndex = traverseReadonlyArrayWithIndex;
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverse(Apply)`.
 *
 * @since 3.0.0
 */
var traverseReadonlyNonEmptyArray = function (f) {
    return (0, exports.traverseReadonlyNonEmptyArrayWithIndex)(function (_, a) { return f(a); });
};
exports.traverseReadonlyNonEmptyArray = traverseReadonlyNonEmptyArray;
/**
 * Equivalent to `ReadonlyArray#traverse(Applicative)`.
 *
 * @since 3.0.0
 */
var traverseReadonlyArray = function (f) {
    return (0, exports.traverseReadonlyArrayWithIndex)(function (_, a) { return f(a); });
};
exports.traverseReadonlyArray = traverseReadonlyArray;
/**
 * Equivalent to `ReadonlyArray#sequence(Applicative)`.
 *
 * @since 3.0.0
 */
exports.sequenceReadonlyArray = 
/*#__PURE__*/ (0, exports.traverseReadonlyArray)(function_1.identity);
