"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.apT = exports.tupled = exports.ApT = exports.apS = exports.bind = exports.let = exports.bindTo = exports.Do = exports.drawTree = exports.drawForest = exports.elem = exports.exists = exports.Comonad = exports.sequence = exports.Traversable = exports.Foldable = exports.chainFirst = exports.Monad = exports.Chain = exports.Applicative = exports.apSecond = exports.apFirst = exports.Apply = exports.Pointed = exports.flap = exports.Functor = exports.getEq = exports.getShow = exports.of = exports.traverse = exports.extract = exports.reduceRight = exports.foldMap = exports.reduce = exports.flatten = exports.duplicate = exports.extend = exports.chain = exports.ap = exports.map = exports.fold = exports.unfoldForestM = exports.unfoldTreeM = exports.unfoldForest = exports.unfoldTree = exports.tree = void 0;
var Apply_1 = require("./Apply");
var Chain_1 = require("./Chain");
var Eq_1 = require("./Eq");
var function_1 = require("./function");
var Functor_1 = require("./Functor");
var _ = __importStar(require("./internal"));
var RA = __importStar(require("./ReadonlyArray"));
var TraversableModule = __importStar(require("./Traversable"));
// -------------------------------------------------------------------------------------
// constructors
// -------------------------------------------------------------------------------------
/**
 * @category constructors
 * @since 3.0.0
 */
var tree = function (value, forest) {
    if (forest === void 0) { forest = RA.empty; }
    return ({
        value: value,
        forest: forest
    });
};
exports.tree = tree;
/**
 * Build a (possibly infinite) tree from a seed value in breadth-first order.
 *
 * @category constructors
 * @since 3.0.0
 */
var unfoldTree = function (f) {
    return function (b) {
        var _a = f(b), a = _a[0], bs = _a[1];
        return {
            value: a,
            forest: (0, function_1.pipe)(bs, (0, exports.unfoldForest)(f))
        };
    };
};
exports.unfoldTree = unfoldTree;
/**
 * Build a (possibly infinite) forest from a list of seed values in breadth-first order.
 *
 * @category constructors
 * @since 3.0.0
 */
var unfoldForest = function (f) {
    return function (bs) {
        return bs.map((0, exports.unfoldTree)(f));
    };
};
exports.unfoldForest = unfoldForest;
/**
 * Monadic tree builder, in depth-first order.
 *
 * @category constructors
 * @since 3.0.0
 */
function unfoldTreeM(M, A) {
    var unfoldForestMM = unfoldForestM(M, A);
    return function (f) {
        return (0, function_1.flow)(f, M.chain(function (_a) {
            var value = _a[0], bs = _a[1];
            return (0, function_1.pipe)(bs, unfoldForestMM(f), M.map(function (forest) { return ({ value: value, forest: forest }); }));
        }));
    };
}
exports.unfoldTreeM = unfoldTreeM;
/**
 * Monadic forest builder, in depth-first order.
 *
 * @category constructors
 * @since 3.0.0
 */
function unfoldForestM(M, A) {
    var traverseM = RA.traverse(A);
    return function (f) { return traverseM(unfoldTreeM(M, A)(f)); };
}
exports.unfoldForestM = unfoldForestM;
// -------------------------------------------------------------------------------------
// destructors
// -------------------------------------------------------------------------------------
/**
 * Fold a tree into a "summary" value in depth-first order.
 *
 * For each node in the tree, apply `f` to the `value` and the result of applying `f` to each `forest`.
 *
 * This is also known as the catamorphism on trees.
 *
 * @example
 * import { fold, tree } from 'fp-ts/Tree'
 * import * as N from 'fp-ts/number'
 * import { concatAll } from 'fp-ts/Monoid'
 * import { pipe } from 'fp-ts/function'
 * import { isEmpty } from 'fp-ts/ReadonlyArray'
 *
 * const t = tree(1, [tree(2), tree(3)])
 *
 * const sum = concatAll(N.MonoidSum)
 *
 * assert.deepStrictEqual(pipe(t, fold((a, bs) => a + sum(bs))), 6)
 * assert.deepStrictEqual(pipe(t, fold((a, bs) => bs.reduce((b, acc) => Math.max(b, acc), a))), 3)
 * assert.deepStrictEqual(pipe(t, fold((_, bs) => (isEmpty(bs) ? 1 : sum(bs)))), 2)
 *
 * @category destructors
 * @since 3.0.0
 */
var fold = function (f) {
    var go = function (tree) { return f(tree.value, tree.forest.map(go)); };
    return go;
};
exports.fold = fold;
// -------------------------------------------------------------------------------------
// type class members
// -------------------------------------------------------------------------------------
/**
 * `map` can be used to turn functions `(a: A) => B` into functions `(fa: F<A>) => F<B>` whose argument and return types
 * use the type constructor `F` to represent some computational context.
 *
 * @category Functor
 * @since 3.0.0
 */
var map = function (f) { return function (fa) { return ({
    value: f(fa.value),
    forest: fa.forest.map((0, exports.map)(f))
}); }; };
exports.map = map;
/**
 * Apply a function to an argument under a type constructor.
 *
 * @category Apply
 * @since 3.0.0
 */
var ap = function (fa) { return (0, exports.chain)(function (f) { return (0, function_1.pipe)(fa, (0, exports.map)(f)); }); };
exports.ap = ap;
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation.
 *
 * @category Chain
 * @since 3.0.0
 */
var chain = function (f) {
    return function (ma) {
        var _a = f(ma.value), value = _a.value, forest = _a.forest;
        var concat = RA.getMonoid().concat;
        return {
            value: value,
            forest: concat(ma.forest.map((0, exports.chain)(f)))(forest)
        };
    };
};
exports.chain = chain;
/**
 * @category Extend
 * @since 3.0.0
 */
var extend = function (f) { return function (wa) { return ({
    value: f(wa),
    forest: wa.forest.map((0, exports.extend)(f))
}); }; };
exports.extend = extend;
/**
 * Derivable from `Extend`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
exports.duplicate = (0, exports.extend)(function_1.identity);
/**
 * Derivable from `Chain`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
exports.flatten = (0, exports.chain)(function_1.identity);
/**
 * @category Foldable
 * @since 3.0.0
 */
var reduce = function (b, f) {
    return function (fa) {
        var r = f(b, fa.value);
        var len = fa.forest.length;
        for (var i = 0; i < len; i++) {
            r = (0, function_1.pipe)(fa.forest[i], (0, exports.reduce)(r, f));
        }
        return r;
    };
};
exports.reduce = reduce;
/**
 * @category Foldable
 * @since 3.0.0
 */
var foldMap = function (M) { return function (f) {
    return (0, exports.reduce)(M.empty, function (acc, a) { return M.concat(f(a))(acc); });
}; };
exports.foldMap = foldMap;
/**
 * @category Foldable
 * @since 3.0.0
 */
var reduceRight = function (b, f) {
    return function (fa) {
        var r = b;
        var len = fa.forest.length;
        for (var i = len - 1; i >= 0; i--) {
            r = (0, function_1.pipe)(fa.forest[i], (0, exports.reduceRight)(r, f));
        }
        return f(fa.value, r);
    };
};
exports.reduceRight = reduceRight;
/**
 * @category Extract
 * @since 3.0.0
 */
var extract = function (wa) { return wa.value; };
exports.extract = extract;
/**
 * @category Traversable
 * @since 3.0.0
 */
var traverse = function (F) {
    var traverseF = RA.traverse(F);
    var out = function (f) {
        return function (ta) {
            return (0, function_1.pipe)(f(ta.value), F.map(function (value) { return function (forest) { return ({
                value: value,
                forest: forest
            }); }; }), F.ap((0, function_1.pipe)(ta.forest, traverseF(out(f)))));
        };
    };
    return out;
};
exports.traverse = traverse;
/**
 * @category Pointed
 * @since 3.0.0
 */
var of = function (a) { return (0, exports.tree)(a); };
exports.of = of;
// -------------------------------------------------------------------------------------
// instances
// -------------------------------------------------------------------------------------
/**
 * @category instances
 * @since 3.0.0
 */
var getShow = function (S) {
    var show = function (t) {
        return RA.isEmpty(t.forest)
            ? "tree(".concat(S.show(t.value), ")")
            : "tree(".concat(S.show(t.value), ", [").concat(t.forest.map(show).join(', '), "])");
    };
    return {
        show: show
    };
};
exports.getShow = getShow;
/**
 * @category instances
 * @since 3.0.0
 */
var getEq = function (E) {
    var R = (0, Eq_1.fromEquals)(function (second) { return function (first) { return E.equals(second.value)(first.value) && SA.equals(second.forest)(first.forest); }; });
    var SA = RA.getEq(R);
    return R;
};
exports.getEq = getEq;
/**
 * @category instances
 * @since 3.0.0
 */
exports.Functor = {
    map: exports.map
};
/**
 * Derivable from `Functor`.
 *
 * @category combinators
 * @since 3.0.0
 */
exports.flap = (0, Functor_1.flap)(exports.Functor);
/**
 * @category instances
 * @since 3.0.0
 */
exports.Pointed = {
    of: exports.of
};
/**
 * @category instances
 * @since 3.0.0
 */
exports.Apply = {
    map: exports.map,
    ap: exports.ap
};
/**
 * Combine two effectful actions, keeping only the result of the first.
 *
 * Derivable from `Apply`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
exports.apFirst = (0, Apply_1.apFirst)(exports.Apply);
/**
 * Combine two effectful actions, keeping only the result of the second.
 *
 * Derivable from `Apply`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
exports.apSecond = (0, Apply_1.apSecond)(exports.Apply);
/**
 * @category instances
 * @since 3.0.0
 */
exports.Applicative = {
    map: exports.map,
    ap: exports.ap,
    of: exports.of
};
/**
 * @category instances
 * @since 3.0.0
 */
exports.Chain = {
    map: exports.map,
    chain: exports.chain
};
/**
 * @category instances
 * @since 3.0.0
 */
exports.Monad = {
    map: exports.map,
    of: exports.of,
    chain: exports.chain
};
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation and
 * keeping only the result of the first.
 *
 * Derivable from `Chain`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
exports.chainFirst = (0, Chain_1.chainFirst)(exports.Chain);
/**
 * @category instances
 * @since 3.0.0
 */
exports.Foldable = {
    reduce: exports.reduce,
    foldMap: exports.foldMap,
    reduceRight: exports.reduceRight
};
/**
 * @category instances
 * @since 3.0.0
 */
exports.Traversable = {
    traverse: exports.traverse
};
/**
 * @since 3.0.0
 */
exports.sequence = 
/*#__PURE__*/ TraversableModule.sequence(exports.Traversable);
/**
 * @category instances
 * @since 3.0.0
 */
exports.Comonad = {
    map: exports.map,
    extend: exports.extend,
    extract: exports.extract
};
// -------------------------------------------------------------------------------------
// utils
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
var exists = function (predicate) {
    return function (ma) {
        return predicate(ma.value) || ma.forest.some((0, exports.exists)(predicate));
    };
};
exports.exists = exists;
/**
 * Tests whether a value is a member of a `Tree`.
 *
 * @since 3.0.0
 */
var elem = function (E) {
    return function (a) {
        return (0, exports.exists)(E.equals(a));
    };
};
exports.elem = elem;
var draw = function (indentation, forest) {
    var r = '';
    var len = forest.length;
    var tree;
    for (var i = 0; i < len; i++) {
        tree = forest[i];
        var isLast = i === len - 1;
        r += indentation + (isLast ? '└' : '├') + '─ ' + tree.value;
        r += draw(indentation + (len > 1 && !isLast ? '│  ' : '   '), tree.forest);
    }
    return r;
};
/**
 * Neat 2-dimensional drawing of a forest
 *
 * @since 3.0.0
 */
var drawForest = function (forest) { return draw('\n', forest); };
exports.drawForest = drawForest;
/**
 * Neat 2-dimensional drawing of a tree
 *
 * @example
 * import { tree, drawTree } from 'fp-ts/Tree'
 *
 * const fa = tree('a', [
 *   tree('b'),
 *   tree('c'),
 *   tree('d', [tree('e'), tree('f')])
 * ])
 *
 * assert.strictEqual(drawTree(fa), `a
 * ├─ b
 * ├─ c
 * └─ d
 *    ├─ e
 *    └─ f`)
 *
 *
 * @since 3.0.0
 */
var drawTree = function (tree) { return tree.value + (0, exports.drawForest)(tree.forest); };
exports.drawTree = drawTree;
// -------------------------------------------------------------------------------------
// do notation
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
exports.Do = (0, exports.of)(_.emptyRecord);
/**
 * @since 3.0.0
 */
exports.bindTo = 
/*#__PURE__*/ (0, Functor_1.bindTo)(exports.Functor);
var let_ = /*#__PURE__*/ (0, Functor_1.let)(exports.Functor);
exports.let = let_;
/**
 * @since 3.0.0
 */
exports.bind = (0, Chain_1.bind)(exports.Chain);
// -------------------------------------------------------------------------------------
// sequence S
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
exports.apS = (0, Apply_1.apS)(exports.Apply);
// -------------------------------------------------------------------------------------
// sequence T
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
exports.ApT = (0, exports.of)(_.emptyReadonlyArray);
/**
 * @since 3.0.0
 */
exports.tupled = (0, Functor_1.tupled)(exports.Functor);
/**
 * @since 3.0.0
 */
exports.apT = 
/*#__PURE__*/ (0, Apply_1.apT)(exports.Apply);
