import { constVoid, flow, pipe } from './function';
import * as FunctorModule from './Functor';
import * as SeparatedModule from './Separated';
import * as _ from './internal';
// -------------------------------------------------------------------------------------
// defaults
// -------------------------------------------------------------------------------------
/**
 * Return a default `compact` implementation from `Functor` and `separate`.
 *
 * @category defaults
 * @since 3.0.0
 */
export const compactDefault = (F) => (separate) => {
    return flow(F.map(_.fromOption(constVoid)), separate, SeparatedModule.right);
};
/**
 * Return a default `separate` implementation from `Functor` and `compact`.
 *
 * @category defaults
 * @since 3.0.0
 */
export function separateDefault(F) {
    return (compact) => (fe) => SeparatedModule.separated(pipe(fe, F.map(_.getLeft), compact), pipe(fe, F.map(_.getRight), compact));
}
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * `compact` composition.
 *
 * @category combinators
 * @since 3.0.0
 */
export function compact(F, G) {
    return F.map(G.compact);
}
/**
 * `separate` composition.
 *
 * @category combinators
 * @since 3.0.0
 */
export function separate(F, C, G) {
    const compactFC = compact(F, C);
    const mapFG = FunctorModule.map(F, G);
    return (fge) => SeparatedModule.separated(pipe(fge, mapFG(_.getLeft), compactFC), pipe(fge, mapFG(_.getRight), compactFC));
}
