/**
 * @since 3.0.0
 */
import * as ApplyModule from './Apply';
import * as EitherModule from './Either';
import { flow, pipe } from './function';
import * as FunctorModule from './Functor';
// -------------------------------------------------------------------------------------
// constructors
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const right = (F) => (a) => F.of(EitherModule.right(a));
/**
 * @since 3.0.0
 */
export const left = (F) => (e) => F.of(EitherModule.left(e));
/**
 * @since 3.0.0
 */
export function rightF(F) {
    return F.map(EitherModule.right);
}
/**
 * @since 3.0.0
 */
export function leftF(F) {
    return F.map(EitherModule.left);
}
// -------------------------------------------------------------------------------------
// type class members
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const map = (F) => FunctorModule.map(F, EitherModule.Functor);
/**
 * @since 3.0.0
 */
export const ap = (F) => {
    return ApplyModule.ap(F, EitherModule.Apply);
};
/**
 * @since 3.0.0
 */
export const chain = (M) => (f) => (ma) => {
    return pipe(ma, M.chain((e) => EitherModule.isLeft(e) ? M.of(e) : f(e.right)));
};
/**
 * @since 3.0.0
 */
export const alt = (M) => (second) => (first) => {
    return pipe(first, M.chain((e) => EitherModule.isLeft(e) ? second() : M.of(e)));
};
/**
 * @since 3.0.0
 */
export function bimap(F) {
    return flow(EitherModule.bimap, F.map);
}
/**
 * @since 3.0.0
 */
export function mapLeft(F) {
    return (f) => F.map(EitherModule.mapLeft(f));
}
/**
 * @since 3.0.0
 */
export const altValidation = (M, S) => (second) => (first) => {
    const rightM = right(M);
    return pipe(first, M.chain(EitherModule.match((e1) => pipe(second(), M.map(EitherModule.mapLeft((e2) => S.concat(e2)(e1)))), rightM)));
};
// -------------------------------------------------------------------------------------
// destructors
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export function match(F) {
    return flow(EitherModule.match, F.map);
}
/**
 * @since 3.0.0
 */
export const matchE = (M) => (onLeft, onRight) => {
    return M.chain(EitherModule.match(onLeft, onRight));
};
/**
 * @since 3.0.0
 */
export const getOrElse = (F) => (onLeft) => {
    return F.map(EitherModule.getOrElse(onLeft));
};
/**
 * @since 3.0.0
 */
export const getOrElseE = (M) => (onLeft) => (ma) => {
    return pipe(ma, M.chain(EitherModule.match(onLeft, M.of)));
};
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const orElse = (M) => (onLeft) => (ma) => {
    return pipe(ma, M.chain((e) => EitherModule.isLeft(e) ? onLeft(e.left) : M.of(e)));
};
/**
 * @since 3.0.0
 */
export const orLeft = (M) => (onLeft) => (fa) => {
    return pipe(fa, M.chain(EitherModule.match((e) => pipe(onLeft(e), M.map(EitherModule.left)), (a) => M.of(EitherModule.right(a)))));
};
/**
 * @since 3.0.0
 */
export const orElseFirst = (M) => {
    const orElseM = orElse(M);
    return (onLeft) => (ma) => {
        return pipe(ma, orElseM((e) => pipe(onLeft(e), M.map((eb) => (EitherModule.isLeft(eb) ? eb : EitherModule.left(e))))));
    };
};
// -------------------------------------------------------------------------------------
// utils
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export function swap(F) {
    return F.map(EitherModule.swap);
}
/**
 * @since 3.0.0
 */
export function toUnion(F) {
    return F.map(EitherModule.toUnion);
}
/**
 * @since 3.0.0
 */
export const bracket = (M) => (acquire, use, release) => {
    const leftM = left(M);
    return pipe(acquire, M.chain(EitherModule.match(leftM, (a) => pipe(use(a), M.chain((e) => pipe(release(a, e), M.chain(EitherModule.match(leftM, () => M.of(e)))))))));
};
