import { flow, pipe } from './function';
import { separated } from './Separated';
import * as _ from './internal';
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * @category combinators
 * @since 3.0.0
 */
export const filter = (F) => (predicate) => F.filterMap((b) => (predicate(b) ? _.some(b) : _.none));
/**
 * @category combinators
 * @since 3.0.0
 */
export const refine = (F) => (refinement) => F.filterMap((b) => (refinement(b) ? _.some(b) : _.none));
/**
 * @category combinators
 * @since 3.0.0
 */
export const partition = (F) => (predicate) => F.partitionMap((b) => (predicate(b) ? _.right(b) : _.left(b)));
/**
 * @category combinators
 * @since 3.0.0
 */
export const refinement = (F) => (refinement) => F.partitionMap((c) => (refinement(c) ? _.right(c) : _.left(c)));
/**
 * `filterMap` composition.
 *
 * @category combinators
 * @since 3.0.0
 */
export const filterMap = (F, G) => {
    return (f) => F.map(G.filterMap(f));
};
/**
 * `partitionMap` composition.
 *
 * @category combinators
 * @since 3.0.0
 */
export const partitionMap = (F, G) => {
    const _filterMap = filterMap(F, G);
    return (f) => (fga) => separated(pipe(fga, _filterMap(flow(f, _.getLeft))), pipe(fga, _filterMap(flow(f, _.getRight))));
};
