import { pipe } from './function';
import * as _ from './internal';
// -------------------------------------------------------------------------------------
// constructors
// -------------------------------------------------------------------------------------
/**
 * @category constructors
 * @since 3.0.0
 */
export const fromPredicate = (F) => (predicate) => (b) => F.fromOption(predicate(b) ? _.some(b) : _.none);
/**
 * @category constructors
 * @since 3.0.0
 */
export const fromRefinement = (F) => (refinement) => (c) => F.fromOption(refinement(c) ? _.some(c) : _.none);
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromOptionK = (F) => (f) => (...a) => F.fromOption(f(...a));
// -------------------------------------------------------------------------------------
// interop
// -------------------------------------------------------------------------------------
/**
 * @category interop
 * @since 3.0.0
 */
export const fromNullable = (F) => (a) => F.fromOption(_.fromNullable(a));
/**
 * @category interop
 * @since 3.0.0
 */
export const fromNullableK = (F) => {
    const fromNullableF = fromNullable(F);
    return (f) => (...a) => {
        return fromNullableF(f(...a));
    };
};
/**
 * @category interop
 * @since 3.0.0
 */
export const chainNullableK = (F, C) => {
    const fromNullableKF = fromNullableK(F);
    return (f) => (ma) => {
        return pipe(ma, C.chain(fromNullableKF(f)));
    };
};
