/**
 * Lift a computation from the `Reader` monad.
 *
 * @since 3.0.0
 */
import * as ChainModule from './Chain';
import { pipe } from './function';
import * as _ from './internal';
// -------------------------------------------------------------------------------------
// constructors
// -------------------------------------------------------------------------------------
/**
 * @category constructors
 * @since 3.0.0
 */
export function ask(F) {
    return () => F.fromReader(_.ask());
}
/**
 * @category constructors
 * @since 3.0.0
 */
export function asks(F) {
    return F.fromReader;
}
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromReaderK = (F) => (f) => (...a) => F.fromReader(f(...a));
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainReaderK = (F, M) => (f) => (ma) => {
    return pipe(ma, M.chain((a) => F.fromReader(f(a))));
};
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainFirstReaderK = (F, M) => {
    const chainFirstM = ChainModule.chainFirst(M);
    return (f) => chainFirstM((a) => F.fromReader(f(a)));
};
