/**
 * A `Functor` is a type constructor which supports a mapping operation `map`.
 *
 * `map` can be used to turn functions `A -> B` into functions `F<A> -> F<B>` whose argument and return types use the type
 * constructor `F` to represent some computational context.
 *
 * Instances must satisfy the following laws:
 *
 * 1. Identity: `map(identity) <-> identity`
 * 2. Composition: `map(flow(ab, bc)) <-> flow(map(ab), map(bc))`
 *
 * @since 3.0.0
 */
import { apply } from './function';
import * as TupleModule from './tuple';
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * @category combinators
 * @since 3.0.0
 */
export const flap = (F) => (a) => F.map(apply(a));
/**
 * `map` composition.
 *
 * @category combinators
 * @since 3.0.0
 */
export const map = (F, G) => (f) => F.map(G.map(f));
// -------------------------------------------------------------------------------------
// utils
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const bindTo = (F) => (name) => F.map((a) => ({ [name]: a }));
/**
 * @since 3.0.0
 */
export const tupled = (F) => F.map(TupleModule.tuple);
const let_ = (F) => {
    return (name, f) => F.map((a) => Object.assign({}, a, { [name]: f(a) }));
};
export { 
/**
 * @since 3.0.0
 */
let_ as let };
