import * as ApplyModule from './Apply';
import * as ChainModule from './Chain';
import * as CompactableModule from './Compactable';
import * as EitherModule from './Either';
import * as EitherTModule from './EitherT';
import * as FilterableModule from './Filterable';
import * as FromEitherModule from './FromEither';
import * as FromIOModule from './FromIO';
import { flow, identity, SK } from './function';
import * as FunctorModule from './Functor';
import * as _ from './internal';
import * as IOModule from './IO';
// -------------------------------------------------------------------------------------
// constructors
// -------------------------------------------------------------------------------------
/**
 * @category constructors
 * @since 3.0.0
 */
export const left = /*#__PURE__*/ EitherTModule.left(IOModule.Pointed);
/**
 * @category constructors
 * @since 3.0.0
 */
export const right = /*#__PURE__*/ EitherTModule.right(IOModule.Pointed);
/**
 * @category constructors
 * @since 3.0.0
 */
export const rightIO = /*#__PURE__*/ EitherTModule.rightF(IOModule.Functor);
/**
 * @category constructors
 * @since 3.0.0
 */
export const leftIO = /*#__PURE__*/ EitherTModule.leftF(IOModule.Functor);
// -------------------------------------------------------------------------------------
// natural transformations
// -------------------------------------------------------------------------------------
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromEither = IOModule.of;
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromIO = rightIO;
// -------------------------------------------------------------------------------------
// destructors
// -------------------------------------------------------------------------------------
/**
 * @category destructors
 * @since 3.0.0
 */
export const match = 
/*#__PURE__*/ EitherTModule.match(IOModule.Functor);
/**
 * @category destructors
 * @since 3.0.0
 */
export const matchE = /*#__PURE__*/ EitherTModule.matchE(IOModule.Monad);
/**
 * @category destructors
 * @since 3.0.0
 */
export const getOrElse = 
/*#__PURE__*/ EitherTModule.getOrElse(IOModule.Functor);
/**
 * @category destructors
 * @since 3.0.0
 */
export const getOrElseE = 
/*#__PURE__*/ EitherTModule.getOrElseE(IOModule.Monad);
// -------------------------------------------------------------------------------------
// interop
// -------------------------------------------------------------------------------------
/**
 * Constructs a new `IOEither` from a function that performs a side effect and might throw.
 *
 * See also [`tryCatchK`](#tryCatchK).
 *
 * @category interop
 * @since 3.0.0
 */
export const tryCatch = (f, onThrow) => () => EitherModule.tryCatch(f, onThrow);
/**
 * Converts a function that may throw to one returning a `IOEither`.
 *
 * @category interop
 * @since 3.0.0
 */
export const tryCatchK = (f, onThrow) => (...a) => tryCatch(() => f(...a), onThrow);
/**
 * @category interop
 * @since 3.0.0
 */
export const toUnion = /*#__PURE__*/ EitherTModule.toUnion(IOModule.Functor);
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * @category combinators
 * @since 3.0.0
 */
export const orElse = /*#__PURE__*/ EitherTModule.orElse(IOModule.Monad);
/**
 * @category combinators
 * @since 3.0.0
 */
export const orElseFirst = /*#__PURE__*/ EitherTModule.orElseFirst(IOModule.Monad);
/**
 * @category combinators
 * @since 3.0.0
 */
export const orElseFirstIOK = (onLeft) => orElseFirst(fromIOK(onLeft));
/**
 * @category combinators
 * @since 3.0.0
 */
export const orLeft = 
/*#__PURE__*/ EitherTModule.orLeft(IOModule.Monad);
/**
 * @category combinators
 * @since 3.0.0
 */
export const swap = /*#__PURE__*/ EitherTModule.swap(IOModule.Functor);
// -------------------------------------------------------------------------------------
// type class members
// -------------------------------------------------------------------------------------
/**
 * `map` can be used to turn functions `(a: A) => B` into functions `(fa: F<A>) => F<B>` whose argument and return types
 * use the type constructor `F` to represent some computational context.
 *
 * @category Functor
 * @since 3.0.0
 */
export const map = /*#__PURE__*/ EitherTModule.map(IOModule.Functor);
/**
 * Map a pair of functions over the two type arguments of the bifunctor.
 *
 * @category Bifunctor
 * @since 3.0.0
 */
export const bimap = 
/*#__PURE__*/ EitherTModule.bimap(IOModule.Functor);
/**
 * Map a function over the first type argument of a bifunctor.
 *
 * @category Bifunctor
 * @since 3.0.0
 */
export const mapLeft = 
/*#__PURE__*/ EitherTModule.mapLeft(IOModule.Functor);
/**
 * Apply a function to an argument under a type constructor.
 *
 * @category Apply
 * @since 3.0.0
 */
export const ap = 
/*#__PURE__*/ EitherTModule.ap(IOModule.Apply);
/**
 * @category Pointed
 * @since 3.0.0
 */
export const of = right;
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation.
 *
 * @category Chain
 * @since 3.0.0
 */
export const chain = 
/*#__PURE__*/ EitherTModule.chain(IOModule.Monad);
/**
 * Derivable from `Chain`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export const flatten = 
/*#__PURE__*/ chain(identity);
/**
 * Identifies an associative operation on a type constructor. It is similar to `Semigroup`, except that it applies to
 * types of kind `* -> *`.
 *
 * @category Alt
 * @since 3.0.0
 */
export const alt = 
/*#__PURE__*/ EitherTModule.alt(IOModule.Monad);
// -------------------------------------------------------------------------------------
// instances
// -------------------------------------------------------------------------------------
/**
 * The default [`ApplicativePar`](#applicativepar) instance returns the first error, if you want to
 * get all errors you need to provide an way to concatenate them via a `Semigroup`.
 *
 * See [`getApplicativeValidation`](./Either.ts.html#getapplicativevalidation).
 *
 * @category instances
 * @since 3.0.0
 */
export const getApplicativeIOValidation = (S) => ({
    map,
    ap: ApplyModule.ap(IOModule.Apply, EitherModule.getApplicativeValidation(S)),
    of
});
/**
 * The default [`Alt`](#alt) instance returns the last error, if you want to
 * get all errors you need to provide an way to concatenate them via a `Semigroup`.
 *
 * See [`getAltValidation`](./Either.ts.html#getaltvalidation).
 *
 * @category instances
 * @since 3.0.0
 */
export const getAltIOValidation = (S) => {
    return {
        map,
        alt: EitherTModule.altValidation(IOModule.Monad, S)
    };
};
/**
 * @category instances
 * @since 3.0.0
 */
export const getCompactable = (M) => {
    const C = EitherModule.getCompactable(M);
    const F = { map: EitherModule.map };
    return {
        compact: CompactableModule.compact(IOModule.Functor, C),
        separate: CompactableModule.separate(IOModule.Functor, C, F)
    };
};
/**
 * @category instances
 * @since 3.0.0
 */
export const getFilterable = (M) => {
    const F = EitherModule.getFilterable(M);
    return {
        filterMap: FilterableModule.filterMap(IOModule.Functor, F),
        partitionMap: FilterableModule.partitionMap(IOModule.Functor, F)
    };
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Functor = {
    map
};
/**
 * Derivable from `Functor`.
 *
 * @category combinators
 * @since 3.0.0
 */
export const flap = 
/*#__PURE__*/ FunctorModule.flap(Functor);
/**
 * @category instances
 * @since 3.0.0
 */
export const Pointed = {
    of
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Bifunctor = {
    bimap,
    mapLeft
};
/**
 * @category instances
 * @since 3.0.0
 */
export const ApplyPar = {
    map,
    ap
};
/**
 * Combine two effectful actions, keeping only the result of the first.
 *
 * Derivable from `Apply`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export const apFirst = 
/*#__PURE__*/ ApplyModule.apFirst(ApplyPar);
/**
 * Combine two effectful actions, keeping only the result of the second.
 *
 * Derivable from `Apply`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export const apSecond = 
/*#__PURE__*/ ApplyModule.apSecond(ApplyPar);
/**
 * @category instances
 * @since 3.0.0
 */
export const ApplicativePar = {
    map,
    ap,
    of
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Chain = {
    map,
    chain
};
const apSeq = /*#__PURE__*/ ChainModule.ap(Chain);
/**
 * @category instances
 * @since 3.0.0
 */
export const ApplySeq = {
    map,
    ap: apSeq
};
/**
 * @category instances
 * @since 3.0.0
 */
export const ApplicativeSeq = {
    map,
    ap: apSeq,
    of
};
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation and
 * keeping only the result of the first.
 *
 * Derivable from `Chain`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export const chainFirst = /*#__PURE__*/ ChainModule.chainFirst(Chain);
/**
 * @category instances
 * @since 3.0.0
 */
export const Monad = {
    map,
    of,
    chain
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Alt = {
    map,
    alt
};
/**
 * @category instances
 * @since 3.0.0
 */
export const FromIO = {
    fromIO
};
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromIOK = /*#__PURE__*/ FromIOModule.fromIOK(FromIO);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainIOK = 
/*#__PURE__*/ FromIOModule.chainIOK(FromIO, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainFirstIOK = 
/*#__PURE__*/ FromIOModule.chainFirstIOK(FromIO, Chain);
/**
 * @category instances
 * @since 3.0.0
 */
export const FromEither = {
    fromEither
};
/**
 * Derivable from `FromEither`.
 *
 * @category natural transformations
 * @since 3.0.0
 */
export const fromOption = 
/*#__PURE__*/ FromEitherModule.fromOption(FromEither);
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromOptionKOrElse = 
/*#__PURE__*/ FromEitherModule.fromOptionKOrElse(FromEither);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainOptionKOrElse = 
/*#__PURE__*/ FromEitherModule.chainOptionKOrElse(FromEither, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainEitherK = 
/*#__PURE__*/ FromEitherModule.chainEitherK(FromEither, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainFirstEitherK = /*#__PURE__*/ FromEitherModule.chainFirstEitherK(FromEither, Chain);
/**
 * Derivable from `FromEither`.
 *
 * @category constructors
 * @since 3.0.0
 */
export const fromPredicateOrElse = /*#__PURE__*/ FromEitherModule.fromPredicateOrElse(FromEither);
/**
 * @category constructors
 * @since 3.0.0
 */
export const fromRefinementOrElse = /*#__PURE__*/ FromEitherModule.fromRefinementOrElse(FromEither);
/**
 * @category combinators
 * @since 3.0.0
 */
export const filterOrElse = /*#__PURE__*/ FromEitherModule.filterOrElse(FromEither, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const refineOrElse = /*#__PURE__*/ FromEitherModule.refineOrElse(FromEither, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromEitherK = /*#__PURE__*/ FromEitherModule.fromEitherK(FromEither);
/**
 * @category interop
 * @since 3.0.0
 */
export const fromNullableOrElse = 
/*#__PURE__*/ FromEitherModule.fromNullableOrElse(FromEither);
/**
 * @category interop
 * @since 3.0.0
 */
export const fromNullableKOrElse = /*#__PURE__*/ FromEitherModule.fromNullableKOrElse(FromEither);
/**
 * @category interop
 * @since 3.0.0
 */
export const chainNullableKOrElse = 
/*#__PURE__*/ FromEitherModule.chainNullableKOrElse(FromEither, Chain);
// -------------------------------------------------------------------------------------
// utils
// -------------------------------------------------------------------------------------
/**
 * Make sure that a resource is cleaned up in the event of an exception (\*). The release action is called regardless of
 * whether the body action throws (\*) or returns.
 *
 * (\*) i.e. returns a `Left`
 *
 * @since 3.0.0
 */
export const bracket = /*#__PURE__*/ EitherTModule.bracket(IOModule.Monad);
// -------------------------------------------------------------------------------------
// do notation
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const Do = /*#__PURE__*/ of(_.emptyRecord);
/**
 * @since 3.0.0
 */
export const bindTo = /*#__PURE__*/ FunctorModule.bindTo(Functor);
const let_ = 
/*#__PURE__*/ FunctorModule.let(Functor);
export { 
/**
 * @since 3.0.0
 */
let_ as let };
/**
 * @since 3.0.0
 */
export const bind = 
/*#__PURE__*/ ChainModule.bind(Chain);
// -------------------------------------------------------------------------------------
// sequence S
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const apS = 
/*#__PURE__*/ ApplyModule.apS(ApplyPar);
// -------------------------------------------------------------------------------------
// sequence T
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const ApT = /*#__PURE__*/ of(_.emptyReadonlyArray);
/**
 * @since 3.0.0
 */
export const tupled = 
/*#__PURE__*/ FunctorModule.tupled(Functor);
/**
 * @since 3.0.0
 */
export const apT = 
/*#__PURE__*/ ApplyModule.apT(ApplyPar);
// -------------------------------------------------------------------------------------
// array utils
// -------------------------------------------------------------------------------------
// --- Par ---
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverseWithIndex(ApplyPar)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArrayWithIndex = (f) => flow(IOModule.traverseReadonlyNonEmptyArrayWithIndex(f), IOModule.map(EitherModule.traverseReadonlyNonEmptyArrayWithIndex(SK)));
/**
 * Equivalent to `ReadonlyArray#traverseWithIndex(ApplicativePar)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArrayWithIndex = (f) => {
    const g = traverseReadonlyNonEmptyArrayWithIndex(f);
    return (as) => (_.isNonEmpty(as) ? g(as) : ApT);
};
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverse(ApplyPar)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArray = (f) => {
    return traverseReadonlyNonEmptyArrayWithIndex((_, a) => f(a));
};
/**
 * Equivalent to `ReadonlyArray#traverse(ApplicativePar)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArray = (f) => {
    return traverseReadonlyArrayWithIndex((_, a) => f(a));
};
/**
 * Equivalent to `ReadonlyArray#sequence(ApplicativePar)`.
 *
 * @since 3.0.0
 */
export const sequenceReadonlyArray = 
/*#__PURE__*/ traverseReadonlyArray(identity);
// --- Seq ---
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverseWithIndex(ApplySeq)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArrayWithIndexSeq = (f) => (as) => () => {
    const e = f(0, _.head(as))();
    if (_.isLeft(e)) {
        return e;
    }
    const out = [e.right];
    for (let i = 1; i < as.length; i++) {
        const e = f(i, as[i])();
        if (_.isLeft(e)) {
            return e;
        }
        out.push(e.right);
    }
    return _.right(out);
};
/**
 * Equivalent to `ReadonlyArray#traverseWithIndex(ApplicativeSeq)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArrayWithIndexSeq = (f) => {
    const g = traverseReadonlyNonEmptyArrayWithIndexSeq(f);
    return (as) => (_.isNonEmpty(as) ? g(as) : ApT);
};
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverse(ApplySeq)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArraySeq = (f) => {
    return traverseReadonlyNonEmptyArrayWithIndexSeq((_, a) => f(a));
};
/**
 * Equivalent to `ReadonlyArray#traverse(ApplicativeSeq)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArraySeq = (f) => {
    return traverseReadonlyArrayWithIndexSeq((_, a) => f(a));
};
/**
 * Equivalent to `ReadonlyArray#sequence(ApplicativeSeq)`.
 *
 * @since 3.0.0
 */
export const sequenceReadonlyArraySeq = 
/*#__PURE__*/ traverseReadonlyArraySeq(identity);
