import * as ApplyModule from './Apply';
import * as ChainModule from './Chain';
import * as CompactableModule from './Compactable';
import * as FilterableModule from './Filterable';
import * as FromOptionModule from './FromOption';
import * as FromEitherModule from './FromEither';
import * as FromIOModule from './FromIO';
import { flow, identity, SK } from './function';
import * as FunctorModule from './Functor';
import * as _ from './internal';
import * as IOModule from './IO';
import * as OptionModule from './Option';
import * as OptionTModule from './OptionT';
import * as ZeroModule from './Zero';
// -------------------------------------------------------------------------------------
// constructors
// -------------------------------------------------------------------------------------
/**
 * @category constructors
 * @since 3.0.0
 */
export const some = /*#__PURE__*/ OptionTModule.some(IOModule.Pointed);
// -------------------------------------------------------------------------------------
// natural transformations
// -------------------------------------------------------------------------------------
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromOption = IOModule.of;
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromEither = 
/*#__PURE__*/ OptionTModule.fromEither(IOModule.Pointed);
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromIO = /*#__PURE__*/ OptionTModule.fromF(IOModule.Functor);
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromIOEither = /*#__PURE__*/ IOModule.map(OptionModule.fromEither);
// -------------------------------------------------------------------------------------
// destructors
// -------------------------------------------------------------------------------------
/**
 * @category destructors
 * @since 3.0.0
 */
export const match = 
/*#__PURE__*/ OptionTModule.match(IOModule.Functor);
/**
 * @category destructors
 * @since 3.0.0
 */
export const matchE = 
/*#__PURE__*/ OptionTModule.matchE(IOModule.Chain);
/**
 * @category destructors
 * @since 3.0.0
 */
export const getOrElse = 
/*#__PURE__*/ OptionTModule.getOrElse(IOModule.Functor);
/**
 * @category destructors
 * @since 3.0.0
 */
export const getOrElseE = 
/*#__PURE__*/ OptionTModule.getOrElseE(IOModule.Monad);
/**
 * @category destructors
 * @since 3.0.0
 */
export const toUndefined = IOModule.map(OptionModule.toUndefined);
/**
 * @category destructors
 * @since 3.0.0
 */
export const toNullable = IOModule.map(OptionModule.toNullable);
// -------------------------------------------------------------------------------------
// type class members
// -------------------------------------------------------------------------------------
/**
 * `map` can be used to turn functions `(a: A) => B` into functions `(fa: F<A>) => F<B>` whose argument and return types
 * use the type constructor `F` to represent some computational context.
 *
 * @category Functor
 * @since 3.0.0
 */
export const map = /*#__PURE__*/ OptionTModule.map(IOModule.Functor);
/**
 * @category Apply
 * @since 3.0.0
 */
export const ap = 
/*#__PURE__*/ OptionTModule.ap(IOModule.Apply);
/**
 * @category Pointed
 * @since 3.0.0
 */
export const of = some;
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation.
 *
 * @category Chain
 * @since 3.0.0
 */
export const chain = 
/*#__PURE__*/ OptionTModule.chain(IOModule.Monad);
/**
 * Derivable from `Chain`.
 *
 * @category combinators
 * @since 3.0.0
 */
export const flatten = /*#__PURE__*/ chain(identity);
/**
 * @category Alt
 * @since 3.0.0
 */
export const alt = 
/*#__PURE__*/ OptionTModule.alt(IOModule.Monad);
/**
 * @category Zero
 * @since 3.0.0
 */
export const zero = /*#__PURE__*/ OptionTModule.zero(IOModule.Pointed);
/**
 * @category constructors
 * @since 3.0.0
 */
export const none = /*#__PURE__*/ zero();
/**
 * @category Compactable
 * @since 3.0.0
 */
export const compact = 
/*#__PURE__*/ CompactableModule.compact(IOModule.Functor, OptionModule.Compactable);
/**
 * @category Compactable
 * @since 3.0.0
 */
export const separate = 
/*#__PURE__*/ CompactableModule.separate(IOModule.Functor, OptionModule.Compactable, OptionModule.Functor);
/**
 * @category Filterable
 * @since 3.0.0
 */
export const filterMap = 
/*#__PURE__*/ FilterableModule.filterMap(IOModule.Functor, OptionModule.Filterable);
/**
 * @category Filterable
 * @since 3.0.0
 */
export const partitionMap = /*#__PURE__*/ FilterableModule.partitionMap(IOModule.Functor, OptionModule.Filterable);
// -------------------------------------------------------------------------------------
// instances
// -------------------------------------------------------------------------------------
/**
 * @category instances
 * @since 3.0.0
 */
export const Functor = {
    map
};
/**
 * Derivable from `Functor`.
 *
 * @category combinators
 * @since 3.0.0
 */
export const flap = 
/*#__PURE__*/ FunctorModule.flap(Functor);
/**
 * @category instances
 * @since 3.0.0
 */
export const Pointed = {
    of
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Apply = {
    map,
    ap
};
/**
 * Combine two effectful actions, keeping only the result of the first.
 *
 * Derivable from `Apply`.
 *
 * @category combinators
 * @since 3.0.0
 */
export const apFirst = 
/*#__PURE__*/ ApplyModule.apFirst(Apply);
/**
 * Combine two effectful actions, keeping only the result of the second.
 *
 * Derivable from `Apply`.
 *
 * @category combinators
 * @since 3.0.0
 */
export const apSecond = 
/*#__PURE__*/ ApplyModule.apSecond(Apply);
/**
 * @category instances
 * @since 3.0.0
 */
export const Applicative = {
    map,
    ap,
    of
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Chain = {
    map,
    chain
};
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation and
 * keeping only the result of the first.
 *
 * Derivable from `Chain`.
 *
 * @category combinators
 * @since 3.0.0
 */
export const chainFirst = 
/*#__PURE__*/ ChainModule.chainFirst(Chain);
/**
 * @category instances
 * @since 3.0.0
 */
export const Alt = {
    map,
    alt
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Zero = {
    zero
};
/**
 * @category constructors
 * @since 3.0.0
 */
export const guard = /*#__PURE__*/ ZeroModule.guard(Zero, Pointed);
/**
 * @category instances
 * @since 3.0.0
 */
export const Alternative = {
    map,
    alt,
    zero
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Monad = {
    map,
    of,
    chain
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Compactable = {
    compact,
    separate
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Filterable = {
    filterMap,
    partitionMap
};
/**
 * @since 3.0.0
 */
export const filter = 
/*#__PURE__*/ FilterableModule.filter(Filterable);
/**
 * @since 3.0.0
 */
export const refine = /*#__PURE__*/ FilterableModule.refine(Filterable);
/**
 * @since 3.0.0
 */
export const partition = /*#__PURE__*/ FilterableModule.partition(Filterable);
/**
 * @since 3.0.0
 */
export const refinement = /*#__PURE__*/ FilterableModule.refinement(Filterable);
/**
 * @category instances
 * @since 3.0.0
 */
export const FromIO = {
    fromIO
};
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromIOK = /*#__PURE__*/ FromIOModule.fromIOK(FromIO);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainIOK = 
/*#__PURE__*/ FromIOModule.chainIOK(FromIO, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainFirstIOK = 
/*#__PURE__*/ FromIOModule.chainFirstIOK(FromIO, Chain);
/**
 * @category instances
 * @since 3.0.0
 */
export const FromOption = {
    fromOption
};
/**
 * @category constructors
 * @since 3.0.0
 */
export const fromPredicate = 
/*#__PURE__*/ FromOptionModule.fromPredicate(FromOption);
/**
 * @category constructors
 * @since 3.0.0
 */
export const fromRefinement = 
/*#__PURE__*/ FromOptionModule.fromRefinement(FromOption);
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromOptionK = 
/*#__PURE__*/ FromOptionModule.fromOptionK(FromOption);
// -------------------------------------------------------------------------------------
// interop
// -------------------------------------------------------------------------------------
/**
 * @category interop
 * @since 3.0.0
 */
export const fromNullable = 
/*#__PURE__*/ FromOptionModule.fromNullable(FromOption);
/**
 * @category interop
 * @since 3.0.0
 */
export const fromNullableK = /*#__PURE__*/ FromOptionModule.fromNullableK(FromOption);
/**
 * @category interop
 * @since 3.0.0
 */
export const chainNullableK = /*#__PURE__*/ FromOptionModule.chainNullableK(FromOption, Chain);
/**
 * @category instances
 * @since 3.0.0
 */
export const FromEither = {
    fromEither
};
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromEitherK = /*#__PURE__*/ FromEitherModule.fromEitherK(FromEither);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainEitherK = 
/*#__PURE__*/ FromEitherModule.chainEitherK(FromEither, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainFirstEitherK = 
/*#__PURE__*/ FromEitherModule.chainFirstEitherK(FromEither, Chain);
// -------------------------------------------------------------------------------------
// do notation
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const Do = /*#__PURE__*/ of(_.emptyRecord);
/**
 * @since 3.0.0
 */
export const bindTo = 
/*#__PURE__*/ FunctorModule.bindTo(Functor);
const let_ = 
/*#__PURE__*/ FunctorModule.let(Functor);
export { 
/**
 * @since 3.0.0
 */
let_ as let };
/**
 * @since 3.0.0
 */
export const bind = 
/*#__PURE__*/ ChainModule.bind(Chain);
// -------------------------------------------------------------------------------------
// sequence S
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const apS = 
/*#__PURE__*/ ApplyModule.apS(Apply);
// -------------------------------------------------------------------------------------
// sequence T
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const ApT = /*#__PURE__*/ of(_.emptyReadonlyArray);
// -------------------------------------------------------------------------------------
// array utils
// -------------------------------------------------------------------------------------
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverseWithIndex(Apply)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArrayWithIndex = (f) => flow(IOModule.traverseReadonlyNonEmptyArrayWithIndex(f), IOModule.map(OptionModule.traverseReadonlyNonEmptyArrayWithIndex(SK)));
/**
 * Equivalent to `ReadonlyArray#traverseWithIndex(Applicative)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArrayWithIndex = (f) => {
    const g = traverseReadonlyNonEmptyArrayWithIndex(f);
    return (as) => (_.isNonEmpty(as) ? g(as) : ApT);
};
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverse(Apply)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArray = (f) => {
    return traverseReadonlyNonEmptyArrayWithIndex((_, a) => f(a));
};
/**
 * Equivalent to `ReadonlyArray#traverse(Applicative)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArray = (f) => {
    return traverseReadonlyArrayWithIndex((_, a) => f(a));
};
/**
 * Equivalent to `ReadonlyArray#sequence(Applicative)`.
 *
 * @since 3.0.0
 */
export const sequenceReadonlyArray = 
/*#__PURE__*/ traverseReadonlyArray(identity);
