/**
 * @since 3.0.0
 */
import * as ApplyModule from './Apply';
import { constant, flow, pipe } from './function';
import * as FunctorModule from './Functor';
import * as _ from './internal';
import * as OptionModule from './Option';
// -------------------------------------------------------------------------------------
// constructors
// -------------------------------------------------------------------------------------
/**
 * @category constructors
 * @since 3.0.0
 */
export const some = (F) => (a) => F.of(_.some(a));
// -------------------------------------------------------------------------------------
// natural transformations
// -------------------------------------------------------------------------------------
/**
 * @category natural transformations
 * @since 3.0.0
 */
export function fromF(F) {
    return F.map(_.some);
}
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromEither = (F) => (e) => F.of(OptionModule.fromEither(e));
// -------------------------------------------------------------------------------------
// destructors
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export function match(F) {
    return flow(OptionModule.match, F.map);
}
/**
 * @since 3.0.0
 */
export const matchE = (M) => (onNone, onSome) => {
    return M.chain(OptionModule.match(onNone, onSome));
};
/**
 * @since 3.0.0
 */
export const getOrElse = (F) => (onNone) => {
    return F.map(OptionModule.getOrElse(onNone));
};
/**
 * @since 3.0.0
 */
export const getOrElseE = (M) => (onNone) => (ma) => {
    return pipe(ma, M.chain(OptionModule.match(onNone, M.of)));
};
// -------------------------------------------------------------------------------------
// type class members
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export function map(F) {
    return FunctorModule.map(F, OptionModule.Functor);
}
/**
 * @since 3.0.0
 */
export const ap = (F) => {
    return ApplyModule.ap(F, OptionModule.Apply);
};
/**
 * @since 3.0.0
 */
export const chain = (M) => (f) => (ma) => {
    return pipe(ma, M.chain(OptionModule.match(() => zero(M)(), f)));
};
/**
 * @since 3.0.0
 */
export const alt = (M) => (second) => (first) => {
    return pipe(first, M.chain(OptionModule.match(second, some(M))));
};
/**
 * @since 3.0.0
 */
export function zero(F) {
    return constant(F.of(_.none));
}
