import * as ApplyModule from './Apply';
import * as ChainModule from './Chain';
import * as CompactableModule from './Compactable';
import * as EitherModule from './Either';
import * as EitherTModule from './EitherT';
import * as FilterableModule from './Filterable';
import * as FromEitherModule from './FromEither';
import * as FromReaderModule from './FromReader';
import { flow, identity, SK } from './function';
import * as FunctorModule from './Functor';
import * as _ from './internal';
import * as ReaderModule from './Reader';
// -------------------------------------------------------------------------------------
// constructors
// -------------------------------------------------------------------------------------
/**
 * @category constructors
 * @since 3.0.0
 */
export const left = /*#__PURE__*/ EitherTModule.left(ReaderModule.Pointed);
/**
 * @category constructors
 * @since 3.0.0
 */
export const right = /*#__PURE__*/ EitherTModule.right(ReaderModule.Pointed);
/**
 * @category constructors
 * @since 3.0.0
 */
export const rightReader = 
/*#__PURE__*/ EitherTModule.rightF(ReaderModule.Functor);
/**
 * @category constructors
 * @since 3.0.0
 */
export const leftReader = 
/*#__PURE__*/ EitherTModule.leftF(ReaderModule.Functor);
/**
 * @category constructors
 * @since 3.0.0
 */
export const asksReaderEither = ReaderModule.asksReader;
// -------------------------------------------------------------------------------------
// natural transformations
// -------------------------------------------------------------------------------------
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromEither = ReaderModule.of;
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromReader = rightReader;
// -------------------------------------------------------------------------------------
// destructors
// -------------------------------------------------------------------------------------
/**
 * @category destructors
 * @since 3.0.0
 */
export const match = /*#__PURE__*/ EitherTModule.match(ReaderModule.Functor);
/**
 * @category destructors
 * @since 3.0.0
 */
export const matchE = 
/*#__PURE__*/ EitherTModule.matchE(ReaderModule.Monad);
/**
 * @category destructors
 * @since 3.0.0
 */
export const getOrElse = 
/*#__PURE__*/ EitherTModule.getOrElse(ReaderModule.Functor);
/**
 * @category destructors
 * @since 3.0.0
 */
export const getOrElseE = /*#__PURE__*/ EitherTModule.getOrElseE(ReaderModule.Monad);
// -------------------------------------------------------------------------------------
// interop
// -------------------------------------------------------------------------------------
/**
 * @category interop
 * @since 3.0.0
 */
export const toUnion = /*#__PURE__*/ EitherTModule.toUnion(ReaderModule.Functor);
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * Changes the value of the local context during the execution of the action `ma` (similar to `Contravariant`'s
 * `contramap`).
 *
 * @category combinators
 * @since 3.0.0
 */
export const local = ReaderModule.local;
/**
 * @category combinators
 * @since 3.0.0
 */
export const orElse = /*#__PURE__*/ EitherTModule.orElse(ReaderModule.Monad);
/**
 * @category combinators
 * @since 3.0.0
 */
export const orElseFirst = 
/*#__PURE__*/ EitherTModule.orElseFirst(ReaderModule.Monad);
/**
 * @category combinators
 * @since 3.0.0
 */
export const orLeft = /*#__PURE__*/ EitherTModule.orLeft(ReaderModule.Monad);
/**
 * @category combinators
 * @since 3.0.0
 */
export const swap = /*#__PURE__*/ EitherTModule.swap(ReaderModule.Functor);
/**
 * `map` can be used to turn functions `(a: A) => B` into functions `(fa: F<A>) => F<B>` whose argument and return types
 * use the type constructor `F` to represent some computational context.
 *
 * @category Functor
 * @since 3.0.0
 */
export const map = 
/*#__PURE__*/ EitherTModule.map(ReaderModule.Functor);
/**
 * Map a pair of functions over the two last type arguments of the bifunctor.
 *
 * @category Bifunctor
 * @since 3.0.0
 */
export const bimap = /*#__PURE__*/ EitherTModule.bimap(ReaderModule.Functor);
/**
 * Map a function over the second type argument of a bifunctor.
 *
 * @category Bifunctor
 * @since 3.0.0
 */
export const mapLeft = 
/*#__PURE__*/ EitherTModule.mapLeft(ReaderModule.Functor);
/**
 * Apply a function to an argument under a type constructor.
 *
 * @category Apply
 * @since 3.0.0
 */
export const ap = 
/*#__PURE__*/ EitherTModule.ap(ReaderModule.Apply);
/**
 * @category Pointed
 * @since 3.0.0
 */
export const of = right;
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation.
 *
 * @category Chain
 * @since 3.0.0
 */
export const chain = /*#__PURE__*/ EitherTModule.chain(ReaderModule.Monad);
/**
 * Derivable from `Chain`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export const flatten = /*#__PURE__*/ chain(identity);
/**
 * Identifies an associative operation on a type constructor. It is similar to `Semigroup`, except that it applies to
 * types of kind `* -> *`.
 *
 * @category Alt
 * @since 3.0.0
 */
export const alt = /*#__PURE__*/ EitherTModule.alt(ReaderModule.Monad);
// -------------------------------------------------------------------------------------
// instances
// -------------------------------------------------------------------------------------
/**
 * The default [`Applicative`](#applicative) instance returns the first error, if you want to
 * get all errors you need to provide an way to concatenate them via a `Semigroup`.
 *
 * See [`getApplicativeValidation`](./Either.ts.html#getapplicativevalidation).
 *
 * @category instances
 * @since 3.0.0
 */
export const getApplicativeReaderValidation = (S) => ({
    map,
    ap: ApplyModule.ap(ReaderModule.Apply, EitherModule.getApplicativeValidation(S)),
    of
});
/**
 * The default [`Alt`](#alt) instance returns the last error, if you want to
 * get all errors you need to provide an way to concatenate them via a `Semigroup`.
 *
 * See [`getAltValidation`](./Either.ts.html#getaltvalidation).
 *
 * @category instances
 * @since 3.0.0
 */
export const getAltReaderValidation = (S) => {
    return {
        map,
        alt: EitherTModule.altValidation(ReaderModule.Monad, S)
    };
};
/**
 * @category instances
 * @since 3.0.0
 */
export const getCompactable = (M) => {
    const C = EitherModule.getCompactable(M);
    const F = { map: EitherModule.map };
    return {
        compact: CompactableModule.compact(ReaderModule.Functor, C),
        separate: CompactableModule.separate(ReaderModule.Functor, C, F)
    };
};
/**
 * @category instances
 * @since 3.0.0
 */
export const getFilterable = (M) => {
    const F = EitherModule.getFilterable(M);
    return {
        filterMap: FilterableModule.filterMap(ReaderModule.Functor, F),
        partitionMap: FilterableModule.partitionMap(ReaderModule.Functor, F)
    };
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Functor = {
    map
};
/**
 * Derivable from `Functor`.
 *
 * @category combinators
 * @since 3.0.0
 */
export const flap = 
/*#__PURE__*/ FunctorModule.flap(Functor);
/**
 * @category instances
 * @since 3.0.0
 */
export const Pointed = {
    of
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Apply = {
    map,
    ap
};
/**
 * Combine two effectful actions, keeping only the result of the first.
 *
 * Derivable from `Apply`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export const apFirst = 
/*#__PURE__*/ ApplyModule.apFirst(Apply);
/**
 * Combine two effectful actions, keeping only the result of the second.
 *
 * Derivable from `Apply`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export const apSecond = 
/*#__PURE__*/ ApplyModule.apSecond(Apply);
/**
 * @category instances
 * @since 3.0.0
 */
export const Applicative = {
    map,
    ap,
    of
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Chain = {
    map,
    chain
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Monad = {
    map,
    of,
    chain
};
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation and
 * keeping only the result of the first.
 *
 * Derivable from `Chain`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export const chainFirst = 
/*#__PURE__*/ ChainModule.chainFirst(Chain);
/**
 * @category instances
 * @since 3.0.0
 */
export const Bifunctor = {
    bimap,
    mapLeft
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Alt = {
    map,
    alt
};
/**
 * @category instances
 * @since 3.0.0
 */
export const FromReader = {
    fromReader
};
/**
 * Reads the current context.
 *
 * @category constructors
 * @since 3.0.0
 */
export const ask = /*#__PURE__*/ FromReaderModule.ask(FromReader);
/**
 * Projects a value from the global context in a `ReaderEither`.
 *
 * @category constructors
 * @since 3.0.0
 */
export const asks = 
/*#__PURE__*/ FromReaderModule.asks(FromReader);
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromReaderK = /*#__PURE__*/ FromReaderModule.fromReaderK(FromReader);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainReaderK = 
/*#__PURE__*/ FromReaderModule.chainReaderK(FromReader, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainFirstReaderK = 
/*#__PURE__*/ FromReaderModule.chainFirstReaderK(FromReader, Chain);
/**
 * @category instances
 * @since 3.0.0
 */
export const FromEither = {
    fromEither
};
/**
 * Derivable from `FromEither`.
 *
 * @category natural transformations
 * @since 3.0.0
 */
export const fromOption = 
/*#__PURE__*/ FromEitherModule.fromOption(FromEither);
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromOptionKOrElse = /*#__PURE__*/ FromEitherModule.fromOptionKOrElse(FromEither);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainOptionKOrElse = 
/*#__PURE__*/ FromEitherModule.chainOptionKOrElse(FromEither, Chain);
/**
 * Derivable from `FromEither`.
 *
 * @category constructors
 * @since 3.0.0
 */
export const fromPredicateOrElse = /*#__PURE__*/ FromEitherModule.fromPredicateOrElse(FromEither);
/**
 * @category constructors
 * @since 3.0.0
 */
export const fromRefinementOrElse = /*#__PURE__*/ FromEitherModule.fromRefinementOrElse(FromEither);
/**
 * @category combinators
 * @since 3.0.0
 */
export const filterOrElse = /*#__PURE__*/ FromEitherModule.filterOrElse(FromEither, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const refineOrElse = /*#__PURE__*/ FromEitherModule.refineOrElse(FromEither, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromEitherK = /*#__PURE__*/ FromEitherModule.fromEitherK(FromEither);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainEitherK = /*#__PURE__*/ FromEitherModule.chainEitherK(FromEither, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainFirstEitherK = 
/*#__PURE__*/ FromEitherModule.chainFirstEitherK(FromEither, Chain);
/**
 * @category interop
 * @since 3.0.0
 */
export const fromNullableOrElse = 
/*#__PURE__*/ FromEitherModule.fromNullableOrElse(FromEither);
/**
 * @category interop
 * @since 3.0.0
 */
export const fromNullableKOrElse = 
/*#__PURE__*/ FromEitherModule.fromNullableKOrElse(FromEither);
/**
 * @category interop
 * @since 3.0.0
 */
export const chainNullableKOrElse = 
/*#__PURE__*/ FromEitherModule.chainNullableKOrElse(FromEither, Chain);
// -------------------------------------------------------------------------------------
// do notation
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const Do = /*#__PURE__*/ of(_.emptyRecord);
/**
 * @since 3.0.0
 */
export const bindTo = 
/*#__PURE__*/ FunctorModule.bindTo(Functor);
const let_ = 
/*#__PURE__*/ FunctorModule.let(Functor);
export { 
/**
 * @since 3.0.0
 */
let_ as let };
/**
 * @since 3.0.0
 */
export const bind = 
/*#__PURE__*/ ChainModule.bind(Chain);
// -------------------------------------------------------------------------------------
// sequence S
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const apS = 
/*#__PURE__*/ ApplyModule.apS(Apply);
// -------------------------------------------------------------------------------------
// sequence T
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const ApT = /*#__PURE__*/ of(_.emptyReadonlyArray);
/**
 * @since 3.0.0
 */
export const tupled = 
/*#__PURE__*/ FunctorModule.tupled(Functor);
/**
 * @since 3.0.0
 */
export const apT = /*#__PURE__*/ ApplyModule.apT(Apply);
// -------------------------------------------------------------------------------------
// utils
// -------------------------------------------------------------------------------------
/**
 * Make sure that a resource is cleaned up in the event of an exception (\*). The release action is called regardless of
 * whether the body action throws (\*) or returns.
 *
 * (\*) i.e. returns a `Left`
 *
 * @since 3.0.0
 */
export const bracket = /*#__PURE__*/ EitherTModule.bracket(ReaderModule.Monad);
// -------------------------------------------------------------------------------------
// array utils
// -------------------------------------------------------------------------------------
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverseWithIndex(Apply)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArrayWithIndex = (f) => flow(ReaderModule.traverseReadonlyNonEmptyArrayWithIndex(f), ReaderModule.map(EitherModule.traverseReadonlyNonEmptyArrayWithIndex(SK)));
/**
 * Equivalent to `ReadonlyArray#traverseWithIndex(Applicative)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArrayWithIndex = (f) => {
    const g = traverseReadonlyNonEmptyArrayWithIndex(f);
    return (as) => (_.isNonEmpty(as) ? g(as) : ApT);
};
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverse(Apply)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArray = (f) => {
    return traverseReadonlyNonEmptyArrayWithIndex((_, a) => f(a));
};
/**
 * Equivalent to `ReadonlyArray#traverse(Applicative)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArray = (f) => {
    return traverseReadonlyArrayWithIndex((_, a) => f(a));
};
/**
 * Equivalent to `ReadonlyArray#sequence(Applicative)`.
 *
 * @since 3.0.0
 */
export const sequenceReadonlyArray = /*#__PURE__*/ traverseReadonlyArray(identity);
