import * as ApplyModule from './Apply';
import * as ChainModule from './Chain';
import * as FromIOModule from './FromIO';
import * as FromReaderModule from './FromReader';
import { flow, identity, SK } from './function';
import * as FunctorModule from './Functor';
import * as _ from './internal';
import * as I from './IO';
import * as ReaderModule from './Reader';
import * as ReaderTModule from './ReaderT';
// -------------------------------------------------------------------------------------
// natural transformations
// -------------------------------------------------------------------------------------
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromReader = 
/*#__PURE__*/ ReaderTModule.fromReader(I.Pointed);
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromIO = /*#__PURE__*/ ReaderModule.of;
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * Changes the value of the local context during the execution of the action `ma` (similar to `Contravariant`'s
 * `contramap`).
 *
 * @category combinators
 * @since 3.0.0
 */
export const local = ReaderModule.local;
/**
 * Effectfully accesses the environment.
 *
 * @category combinators
 * @since 3.0.0
 */
export const asksReaderIO = ReaderModule.asksReader;
/**
 * `map` can be used to turn functions `(a: A) => B` into functions `(fa: F<A>) => F<B>` whose argument and return types
 * use the type constructor `F` to represent some computational context.
 *
 * @category Functor
 * @since 3.0.0
 */
export const map = /*#__PURE__*/ ReaderTModule.map(I.Functor);
/**
 * Apply a function to an argument under a type constructor.
 *
 * @category Apply
 * @since 3.0.0
 */
export const ap = 
/*#__PURE__*/ ReaderTModule.ap(I.Apply);
/**
 * @category Pointed
 * @since 3.0.0
 */
export const of = /*#__PURE__*/ ReaderTModule.of(I.Pointed);
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation.
 *
 * @category Monad
 * @since 3.0.0
 */
export const chain = 
/*#__PURE__*/ ReaderTModule.chain(I.Monad);
/**
 * Derivable from `Chain`.
 *
 * @category combinators
 * @since 3.0.0
 */
export const flatten = 
/*#__PURE__*/ chain(identity);
// -------------------------------------------------------------------------------------
// instances
// -------------------------------------------------------------------------------------
/**
 * @category instances
 * @since 3.0.0
 */
export const Functor = {
    map
};
/**
 * Derivable from `Functor`.
 *
 * @category combinators
 * @since 3.0.0
 */
export const flap = 
/*#__PURE__*/ FunctorModule.flap(Functor);
/**
 * @category instances
 * @since 3.0.0
 */
export const Pointed = {
    of
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Apply = {
    map,
    ap
};
/**
 * Combine two effectful actions, keeping only the result of the first.
 *
 * Derivable from `Apply`.
 *
 * @category combinators
 * @since 3.0.0
 */
export const apFirst = 
/*#__PURE__*/ ApplyModule.apFirst(Apply);
/**
 * Combine two effectful actions, keeping only the result of the second.
 *
 * Derivable from `Apply`.
 *
 * @category combinators
 * @since 3.0.0
 */
export const apSecond = 
/*#__PURE__*/ ApplyModule.apSecond(Apply);
/**
 * @category instances
 * @since 3.0.0
 */
export const Applicative = {
    map,
    ap,
    of
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Chain = {
    map,
    chain
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Monad = {
    map,
    of,
    chain
};
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation and
 * keeping only the result of the first.
 *
 * Derivable from `Chain`.
 *
 * @category combinators
 * @since 3.0.0
 */
export const chainFirst = 
/*#__PURE__*/ ChainModule.chainFirst(Chain);
/**
 * @category instances
 * @since 3.0.0
 */
export const FromIO = {
    fromIO
};
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromIOK = /*#__PURE__*/ FromIOModule.fromIOK(FromIO);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainIOK = 
/*#__PURE__*/ FromIOModule.chainIOK(FromIO, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainFirstIOK = 
/*#__PURE__*/ FromIOModule.chainFirstIOK(FromIO, Chain);
/**
 * @category instances
 * @since 3.0.0
 */
export const FromReader = {
    fromReader
};
/**
 * Reads the current context.
 *
 * @category constructors
 * @since 3.0.0
 */
export const ask = /*#__PURE__*/ FromReaderModule.ask(FromReader);
/**
 * Projects a value from the global context in a `ReaderIO`.
 *
 * @category constructors
 * @since 3.0.0
 */
export const asks = /*#__PURE__*/ FromReaderModule.asks(FromReader);
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromReaderK = /*#__PURE__*/ FromReaderModule.fromReaderK(FromReader);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainReaderK = /*#__PURE__*/ FromReaderModule.chainReaderK(FromReader, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainFirstReaderK = /*#__PURE__*/ FromReaderModule.chainFirstReaderK(FromReader, Chain);
// -------------------------------------------------------------------------------------
// do notation
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const Do = /*#__PURE__*/ of(_.emptyRecord);
/**
 * @since 3.0.0
 */
export const bindTo = /*#__PURE__*/ FunctorModule.bindTo(Functor);
const let_ = 
/*#__PURE__*/ FunctorModule.let(Functor);
export { 
/**
 * @since 3.0.0
 */
let_ as let };
/**
 * @since 3.0.0
 */
export const bind = 
/*#__PURE__*/ ChainModule.bind(Chain);
// -------------------------------------------------------------------------------------
// pipeable sequence S
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const apS = 
/*#__PURE__*/ ApplyModule.apS(Apply);
// -------------------------------------------------------------------------------------
// sequence T
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const ApT = /*#__PURE__*/ of(_.emptyReadonlyArray);
// -------------------------------------------------------------------------------------
// array utils
// -------------------------------------------------------------------------------------
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverseWithIndex(Apply)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArrayWithIndex = (f) => flow(ReaderModule.traverseReadonlyNonEmptyArrayWithIndex(f), ReaderModule.map(I.traverseReadonlyNonEmptyArrayWithIndex(SK)));
/**
 * Equivalent to `ReadonlyArray#traverseWithIndex(Applicative)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArrayWithIndex = (f) => {
    const g = traverseReadonlyNonEmptyArrayWithIndex(f);
    return (as) => (_.isNonEmpty(as) ? g(as) : ApT);
};
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverse(Apply)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArray = (f) => {
    return traverseReadonlyNonEmptyArrayWithIndex((_, a) => f(a));
};
/**
 * Equivalent to `ReadonlyArray#traverse(Applicative)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArray = (f) => {
    return traverseReadonlyArrayWithIndex((_, a) => f(a));
};
/**
 * Equivalent to `ReadonlyArray#sequence(Applicative)`.
 *
 * @since 3.0.0
 */
export const sequenceReadonlyArray = 
/*#__PURE__*/ traverseReadonlyArray(identity);
