import * as ApplyModule from './Apply';
import * as ChainModule from './Chain';
import * as FromIOModule from './FromIO';
import * as FromReaderModule from './FromReader';
import * as FromTaskModule from './FromTask';
import { flow, identity, SK } from './function';
import * as FunctorModule from './Functor';
import * as _ from './internal';
import * as ReaderModule from './Reader';
import * as ReaderTModule from './ReaderT';
import * as TaskModule from './Task';
// -------------------------------------------------------------------------------------
// constructors
// -------------------------------------------------------------------------------------
/**
 * @category constructors
 * @since 3.0.0
 */
export const asksReaderTask = ReaderModule.asksReader;
// -------------------------------------------------------------------------------------
// natural transformations
// -------------------------------------------------------------------------------------
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromReader = 
/*#__PURE__*/ ReaderTModule.fromReader(TaskModule.Pointed);
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromTask = /*#__PURE__*/ ReaderModule.of;
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromIO = /*#__PURE__*/ flow(TaskModule.fromIO, fromTask);
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromReaderIO = ReaderModule.map(TaskModule.fromIO);
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * Changes the value of the local context during the execution of the action `ma` (similar to `Contravariant`'s
 * `contramap`).
 *
 * @category combinators
 * @since 3.0.0
 */
export const local = ReaderModule.local;
/**
 * `map` can be used to turn functions `(a: A) => B` into functions `(fa: F<A>) => F<B>` whose argument and return types
 * use the type constructor `F` to represent some computational context.
 *
 * @category Functor
 * @since 3.0.0
 */
export const map = 
/*#__PURE__*/ ReaderTModule.map(TaskModule.Functor);
/**
 * Apply a function to an argument under a type constructor.
 *
 * @category Apply
 * @since 3.0.0
 */
export const ap = 
/*#__PURE__*/ ReaderTModule.ap(TaskModule.ApplyPar);
/**
 * @category Pointed
 * @since 3.0.0
 */
export const of = /*#__PURE__*/ ReaderTModule.of(TaskModule.Pointed);
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation.
 *
 * @category Chain
 * @since 3.0.0
 */
export const chain = /*#__PURE__*/ ReaderTModule.chain(TaskModule.Monad);
/**
 * Derivable from `Chain`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export const flatten = 
/*#__PURE__*/ chain(identity);
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromReaderIOK = (f) => (...a) => fromReaderIO(f(...a));
/**
 * Less strict version of [`chainReaderIOK`](#chainreaderiok).
 *
 * @category combinators
 * @since 3.0.0
 */
export const chainReaderIOKW = (f) => chain(fromReaderIOK(f));
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainReaderIOK = chainReaderIOKW;
/**
 * Less strict version of [`chainFirstReaderIOK`](#chainfirstreaderiok).
 *
 * @category combinators
 * @since 3.0.0
 */
export const chainFirstReaderIOKW = (f) => chainFirst(fromReaderIOK(f));
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainFirstReaderIOK = chainFirstReaderIOKW;
// -------------------------------------------------------------------------------------
// instances
// -------------------------------------------------------------------------------------
/**
 * @category instances
 * @since 3.0.0
 */
export const Functor = {
    map
};
/**
 * Derivable from `Functor`.
 *
 * @category combinators
 * @since 3.0.0
 */
export const flap = 
/*#__PURE__*/ FunctorModule.flap(Functor);
/**
 * @category instances
 * @since 3.0.0
 */
export const Pointed = {
    of
};
/**
 * @category instances
 * @since 3.0.0
 */
export const ApplyPar = {
    map,
    ap
};
/**
 * Combine two effectful actions, keeping only the result of the first.
 *
 * Derivable from `Apply`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export const apFirst = 
/*#__PURE__*/ ApplyModule.apFirst(ApplyPar);
/**
 * Combine two effectful actions, keeping only the result of the second.
 *
 * Derivable from `Apply`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export const apSecond = 
/*#__PURE__*/ ApplyModule.apSecond(ApplyPar);
/**
 * @category instances
 * @since 3.0.0
 */
export const ApplicativePar = {
    map,
    ap,
    of
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Chain = {
    map,
    chain
};
const apSeq = /*#__PURE__*/ ChainModule.ap(Chain);
/**
 * @category instances
 * @since 3.0.0
 */
export const ApplySeq = {
    map,
    ap: apSeq
};
/**
 * @category instances
 * @since 3.0.0
 */
export const ApplicativeSeq = {
    map,
    ap: apSeq,
    of
};
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation and
 * keeping only the result of the first.
 *
 * Derivable from `Chain`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export const chainFirst = /*#__PURE__*/ ChainModule.chainFirst(Chain);
/**
 * @category instances
 * @since 3.0.0
 */
export const Monad = {
    map,
    of,
    chain
};
/**
 * @category instances
 * @since 3.0.0
 */
export const FromIO = {
    fromIO
};
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromIOK = /*#__PURE__*/ FromIOModule.fromIOK(FromIO);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainIOK = 
/*#__PURE__*/ FromIOModule.chainIOK(FromIO, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainFirstIOK = 
/*#__PURE__*/ FromIOModule.chainFirstIOK(FromIO, Chain);
/**
 * @category instances
 * @since 3.0.0
 */
export const FromReader = {
    fromReader
};
/**
 * Reads the current context.
 *
 * @category constructors
 * @since 3.0.0
 */
export const ask = /*#__PURE__*/ FromReaderModule.ask(FromReader);
/**
 * Projects a value from the global context in a `ReaderTask`.
 *
 * @category constructors
 * @since 3.0.0
 */
export const asks = /*#__PURE__*/ FromReaderModule.asks(FromReader);
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromReaderK = /*#__PURE__*/ FromReaderModule.fromReaderK(FromReader);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainReaderK = /*#__PURE__*/ FromReaderModule.chainReaderK(FromReader, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainFirstReaderK = /*#__PURE__*/ FromReaderModule.chainFirstReaderK(FromReader, Chain);
/**
 * @category instances
 * @since 3.0.0
 */
export const FromTask = {
    fromIO,
    fromTask
};
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromTaskK = /*#__PURE__*/ FromTaskModule.fromTaskK(FromTask);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainTaskK = 
/*#__PURE__*/ FromTaskModule.chainTaskK(FromTask, Chain);
/**
 * @category combinators
 * @since 3.0.0
 */
export const chainFirstTaskK = /*#__PURE__*/ FromTaskModule.chainFirstTaskK(FromTask, Chain);
// -------------------------------------------------------------------------------------
// do notation
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const Do = /*#__PURE__*/ of(_.emptyRecord);
/**
 * @since 3.0.0
 */
export const bindTo = 
/*#__PURE__*/ FunctorModule.bindTo(Functor);
const let_ = 
/*#__PURE__*/ FunctorModule.let(Functor);
export { 
/**
 * @since 3.0.0
 */
let_ as let };
/**
 * @since 3.0.0
 */
export const bind = 
/*#__PURE__*/ ChainModule.bind(Chain);
// -------------------------------------------------------------------------------------
// sequence S
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const apS = 
/*#__PURE__*/ ApplyModule.apS(ApplyPar);
// -------------------------------------------------------------------------------------
// sequence T
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const ApT = /*#__PURE__*/ of(_.emptyReadonlyArray);
/**
 * @since 3.0.0
 */
export const tupled = 
/*#__PURE__*/ FunctorModule.tupled(Functor);
/**
 * @since 3.0.0
 */
export const apT = 
/*#__PURE__*/ ApplyModule.apT(ApplyPar);
// -------------------------------------------------------------------------------------
// array utils
// -------------------------------------------------------------------------------------
// --- Par ---
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverseWithIndex(ApplyPar)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArrayWithIndex = (f) => flow(ReaderModule.traverseReadonlyNonEmptyArrayWithIndex(f), ReaderModule.map(TaskModule.traverseReadonlyNonEmptyArrayWithIndex(SK)));
/**
 * Equivalent to `ReadonlyArray#traverseWithIndex(ApplicativePar)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArrayWithIndex = (f) => {
    const g = traverseReadonlyNonEmptyArrayWithIndex(f);
    return (as) => (_.isNonEmpty(as) ? g(as) : ApT);
};
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverse(ApplyPar)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArray = (f) => {
    return traverseReadonlyNonEmptyArrayWithIndex((_, a) => f(a));
};
/**
 * Equivalent to `ReadonlyArray#traverse(ApplicativePar)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArray = (f) => {
    return traverseReadonlyArrayWithIndex((_, a) => f(a));
};
/**
 * Equivalent to `ReadonlyArray#sequence(ApplicativePar)`.
 *
 * @since 3.0.0
 */
export const sequenceReadonlyArray = 
/*#__PURE__*/ traverseReadonlyArray(identity);
// --- Seq ---
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverseWithIndex(ApplySeq)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArrayWithIndexSeq = (f) => flow(ReaderModule.traverseReadonlyNonEmptyArrayWithIndex(f), ReaderModule.map(TaskModule.traverseReadonlyNonEmptyArrayWithIndexSeq(SK)));
/**
 * Equivalent to `ReadonlyArray#traverseWithIndex(ApplicativeSeq)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArrayWithIndexSeq = (f) => {
    const g = traverseReadonlyNonEmptyArrayWithIndexSeq(f);
    return (as) => (_.isNonEmpty(as) ? g(as) : ApT);
};
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverse(ApplySeq)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArraySeq = (f) => {
    return traverseReadonlyNonEmptyArrayWithIndexSeq((_, a) => f(a));
};
/**
 * Equivalent to `ReadonlyArray#traverse(ApplicativeSeq)`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArraySeq = (f) => {
    return traverseReadonlyArrayWithIndexSeq((_, a) => f(a));
};
/**
 * Equivalent to `ReadonlyArray#sequence(ApplicativeSeq)`.
 *
 * @since 3.0.0
 */
export const sequenceReadonlyArraySeq = 
/*#__PURE__*/ traverseReadonlyArraySeq(identity);
