import * as FromWriterModule from './FromWriter';
import { flow, identity } from './function';
import * as FunctorModule from './Functor';
import * as _ from './internal';
import * as ReaderModule from './Reader';
import * as ReaderTaskModule from './ReaderTask';
import * as ReadonlyNonEmptyArrayModule from './ReadonlyNonEmptyArray';
import * as TaskModule from './Task';
import * as WriterTModule from './WriterT';
// -------------------------------------------------------------------------------------
// constructors
// -------------------------------------------------------------------------------------
/**
 * @category constructors
 * @since 3.0.0
 */
export const fromReader = (w) => (fa) => fromReaderTask(w)(ReaderTaskModule.fromReader(fa));
/**
 * @category constructors
 * @since 3.0.0
 */
export const fromReaderTask = 
/*#__PURE__*/ WriterTModule.fromF(ReaderTaskModule.Functor);
/**
 * @category constructors
 * @since 3.0.0
 */
export const fromTaskWriter = 
/*#__PURE__*/ ReaderModule.of;
/**
 * @category constructors
 * @since 3.0.0
 */
export const fromIO = 
/*#__PURE__*/ WriterTModule.fromIO(ReaderTaskModule.Functor, ReaderTaskModule.FromIO);
/**
 * @category constructors
 * @since 3.0.0
 */
export const fromTask = 
/*#__PURE__*/ WriterTModule.fromTask(ReaderTaskModule.Functor, ReaderTaskModule.FromTask);
/**
 * Appends a value to the accumulator
 *
 * @category constructors
 * @since 3.0.0
 */
export const tell = /*#__PURE__*/ WriterTModule.tell(ReaderTaskModule.Pointed);
/**
 * @category constructors
 * @since 3.0.0
 */
export const asksReaderTaskWriter = ReaderModule.asksReader;
// -------------------------------------------------------------------------------------
// natural transformations
// -------------------------------------------------------------------------------------
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromWriter = (w) => ReaderTaskModule.of(w);
/**
 * @category natural transformations
 * @since 3.0.0
 */
export const fromReaderWriter = (fa) => flow(fa, TaskModule.of);
// -------------------------------------------------------------------------------------
// utils
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const fst = /*#__PURE__*/ WriterTModule.fst(ReaderTaskModule.Functor);
/**
 * @since 3.0.0
 */
export const snd = /*#__PURE__*/ WriterTModule.snd(ReaderTaskModule.Functor);
/**
 * Alias of [`fst`](#fst).
 *
 * @since 3.0.0
 */
export const evaluate = fst;
/**
 * Alias of [`snd`](#snd).
 *
 * @since 3.0.0
 */
export const execute = snd;
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * Changes the value of the local context during the execution of the action `ma` (similar to `Contravariant`'s
 * `contramap`).
 *
 * @category combinators
 * @since 3.0.0
 */
export const local = ReaderModule.local;
/**
 * @since 3.0.0
 */
export const swap = 
/*#__PURE__*/ WriterTModule.swap(ReaderTaskModule.Functor);
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromTaskWriterK = (f) => (...a) => fromTaskWriter(f(...a));
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromReaderWriterK = (f) => flow(f, fromReaderWriter);
/**
 * @since 3.0.0
 */
export const listen = 
/*#__PURE__*/ WriterTModule.listen(ReaderTaskModule.Functor);
/**
 * @since 3.0.0
 */
export const pass = 
/*#__PURE__*/ WriterTModule.pass(ReaderTaskModule.Functor);
/**
 * @since 3.0.0
 */
export const listens = 
/*#__PURE__*/ WriterTModule.listens(ReaderTaskModule.Functor);
/**
 * @since 3.0.0
 */
export const censor = 
/*#__PURE__*/ WriterTModule.censor(ReaderTaskModule.Functor);
// -------------------------------------------------------------------------------------
// type class operations
// -------------------------------------------------------------------------------------
/**
 * `map` can be used to turn functions `(a: A) => B` into functions `(fa: F<A>) => F<B>` whose argument and return types
 * use the type constructor `F` to represent some computational context.
 *
 * @category type class operations
 * @since 3.0.0
 */
export const map = 
/*#__PURE__*/ WriterTModule.map(ReaderTaskModule.Functor);
/**
 * @category type class operations
 * @since 3.0.0
 */
export const mapLeft = 
/*#__PURE__*/ WriterTModule.mapLeft(ReaderTaskModule.Functor);
/**
 * Map a pair of functions over the two type arguments of the bifunctor.
 *
 * @category type class operations
 * @since 3.0.0
 */
export const bimap = /*#__PURE__*/ WriterTModule.bimap(ReaderTaskModule.Functor);
/**
 * Maps a function over the first component of a `Writer`.
 *
 * Alias of [`map`](#map)
 *
 * @since 3.0.0
 */
export const mapFst = map;
/**
 * Maps a function over the second component of a `Writer`.
 *
 * Alias of [`mapLeft`](#mapleft)
 *
 * @since 3.0.0
 */
export const mapSnd = mapLeft;
// -------------------------------------------------------------------------------------
// instances
// -------------------------------------------------------------------------------------
/**
 * @category instances
 * @since 3.0.0
 */
export const Bifunctor = {
    bimap,
    mapLeft: mapSnd
};
/**
 * @category instances
 * @since 3.0.0
 */
export const Functor = {
    map
};
/**
 * Derivable from `Functor`.
 *
 * @category combinators
 * @since 3.0.0
 */
export const flap = 
/*#__PURE__*/ FunctorModule.flap(Functor);
/**
 * @category instances
 * @since 3.0.0
 */
export const getPointed = (M) => ({
    of: WriterTModule.of(ReaderTaskModule.Pointed, M)
});
/**
 * @category instances
 * @since 3.0.0
 */
export const getApply = (A, S) => ({
    map,
    ap: WriterTModule.ap(A, S)
});
/**
 * @category instances
 * @since 3.0.0
 */
export const getApplicative = (A, M) => {
    const { ap } = getApply(A, M);
    const P = getPointed(M);
    return {
        map,
        ap,
        of: P.of
    };
};
/**
 * @category instances
 * @since 3.0.0
 */
export const getChain = (S) => {
    return {
        map,
        chain: WriterTModule.chain(ReaderTaskModule.Chain, S)
    };
};
/**
 * @category instances
 * @since 3.0.0
 */
export const getMonad = (M) => {
    const P = getPointed(M);
    const C = getChain(M);
    return {
        map,
        of: P.of,
        chain: C.chain
    };
};
/**
 * @category instances
 * @since 3.0.0
 */
export const FromWriter = {
    fromWriter
};
/**
 * @category combinators
 * @since 3.0.0
 */
export const fromWriterK = /*#__PURE__*/ FromWriterModule.fromWriterK(FromWriter);
/**
 * @category instances
 * @since 3.0.0
 */
export const getFromReader = (M) => ({
    fromReader: fromReader(M.empty)
});
/**
 * @category instances
 * @since 3.0.0
 */
export const getFromIO = (M) => ({
    fromIO: fromIO(M.empty)
});
/**
 * @category instances
 * @since 3.0.0
 */
export const getFromTask = (M) => ({
    fromIO: fromIO(M.empty),
    fromTask: fromTask(M.empty)
});
// -------------------------------------------------------------------------------------
// do notation
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const bindTo = 
/*#__PURE__*/ FunctorModule.bindTo(Functor);
const let_ = 
/*#__PURE__*/ FunctorModule.let(Functor);
export { 
/**
 * @since 3.0.0
 */
let_ as let };
// -------------------------------------------------------------------------------------
// sequence T
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export const tupled = 
/*#__PURE__*/ FunctorModule.tupled(Functor);
// -------------------------------------------------------------------------------------
// array utils
// -------------------------------------------------------------------------------------
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverseWithIndex(getApply(A, M))`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArrayWithIndex = (A, S) => (f) => (as) => {
    // TODO
    return ReadonlyNonEmptyArrayModule.traverseWithIndex(getApply(A, S))(f)(as);
};
/**
 * Equivalent to `ReadonlyArray#traverseWithIndex(getApplicative(A, M))`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArrayWithIndex = (A, M) => (f) => {
    const g = traverseReadonlyNonEmptyArrayWithIndex(A, M)(f);
    const P = getPointed(M);
    return (as) => (_.isNonEmpty(as) ? g(as) : P.of(_.emptyReadonlyArray));
};
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverse(getApply(A, M))`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyNonEmptyArray = (A, S) => {
    const traverseReadonlyNonEmptyArrayWithIndexAM = traverseReadonlyNonEmptyArrayWithIndex(A, S);
    return (f) => {
        return traverseReadonlyNonEmptyArrayWithIndexAM((_, a) => f(a));
    };
};
/**
 * Equivalent to `ReadonlyArray#traverse(getApplicative(A, M))`.
 *
 * @since 3.0.0
 */
export const traverseReadonlyArray = (A, M) => {
    const traverseReadonlyArrayWithIndexAM = traverseReadonlyArrayWithIndex(A, M);
    return (f) => {
        return traverseReadonlyArrayWithIndexAM((_, a) => f(a));
    };
};
/**
 * Equivalent to `ReadonlyArray#sequence(getApplicative(A, M))`.
 *
 * @since 3.0.0
 */
export const sequenceReadonlyArray = (A, M) => traverseReadonlyArray(A, M)(identity);
