import { flap as flap_ } from './Functor';
// -------------------------------------------------------------------------------------
// type class members
// -------------------------------------------------------------------------------------
/**
 * `map` can be used to turn functions `(a: A) => B` into functions `(fa: F<A>) => F<B>` whose argument and return types
 * use the type constructor `F` to represent some computational context.
 *
 * @category Functor
 * @since 3.0.0
 */
export const map = (f) => (fa) => (w) => f(fa(w));
// -------------------------------------------------------------------------------------
// instances
// -------------------------------------------------------------------------------------
/**
 * @category instances
 * @since 3.0.0
 */
export const Functor = {
    map
};
/**
 * Derivable from `Functor`.
 *
 * @category combinators
 * @since 3.0.0
 */
export const flap = /*#__PURE__*/ flap_(Functor);
/**
 * @category instances
 * @since 3.0.0
 */
export const getComonad = (monoid) => ({
    map,
    extend: (f) => (fa) => (w1) => f((w2) => fa(monoid.concat(w2)(w1))),
    extract: (fa) => fa(monoid.empty)
});
// -------------------------------------------------------------------------------------
// utils
// -------------------------------------------------------------------------------------
/**
 * Extracts a value at a relative position which depends on the current value.
 *
 * @since 3.0.0
 */
export const tracks = (M) => (f) => (fa) => fa(f(fa(M.empty)));
/**
 * Get the current position.
 *
 * @since 3.0.0
 */
export const listen = (fa) => (w) => [fa(w), w];
/**
 * Get a value which depends on the current position.
 *
 * @since 3.0.0
 */
export const listens = (f) => (pa) => (w) => [pa(w), f(w)];
/**
 * Apply a function to the current position.
 *
 * @since 3.0.0
 */
export const censor = (f) => (pa) => (w) => pa(f(w));
