import { flow, pipe } from './function';
import * as _ from './internal';
// -------------------------------------------------------------------------------------
// defaults
// -------------------------------------------------------------------------------------
/**
 * Return a `wilt` implementation from `Traversable` and `Compactable`.
 *
 * @category defaults
 * @since 3.0.0
 */
export function wiltDefault(T, C) {
    return (F) => {
        const traverseF = T.traverse(F);
        return (f) => {
            return flow(traverseF(f), F.map(C.separate));
        };
    };
}
/**
 * Return a `wither` implementation from `Traversable` and `Compactable`.
 *
 * @category defaults
 * @since 3.0.0
 */
export function witherDefault(T, C) {
    return (F) => {
        const traverseF = T.traverse(F);
        return (f) => {
            return flow(traverseF(f), F.map(C.compact));
        };
    };
}
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * Filter values inside a `F` context.
 *
 * See `ReadonlyArray`'s `filterE` for an example of usage.
 *
 * @since 3.0.0
 */
export const filterE = (F) => (G) => {
    const filterMapEG = F.wither(G);
    return (predicateK) => {
        return filterMapEG((a) => pipe(predicateK(a), G.map((b) => (b ? _.some(a) : _.none))));
    };
};
