/**
 * @since 3.0.0
 */
import { ap as ap_ } from './Apply';
import { pipe } from './function';
import { map as map_ } from './Functor';
import * as W from './Writer';
// -------------------------------------------------------------------------------------
// constructors
// -------------------------------------------------------------------------------------
/**
 * @category constructors
 * @since 3.0.0
 */
export function fromF(F) {
    return (w) => F.map(W.fromIdentity(w));
}
/**
 * @category constructors
 * @since 3.0.0
 */
export const fromIO = (F, FT) => (w) => (fa) => {
    return pipe(FT.fromIO(fa), F.map(W.fromIdentity(w)));
};
/**
 * @category constructors
 * @since 3.0.0
 */
export const fromTask = (F, FT) => (w) => (fa) => {
    return pipe(FT.fromTask(fa), F.map(W.fromIdentity(w)));
};
/**
 * @category constructors
 * @since 3.0.0
 */
export const tell = (F) => (w) => {
    return F.of(W.tell(w));
};
// -------------------------------------------------------------------------------------
// type class operations
// -------------------------------------------------------------------------------------
/**
 * @category type class operations
 * @since 3.0.0
 */
export function map(F) {
    return map_(F, W.Functor);
}
/**
 * @category type class operations
 * @since 3.0.0
 */
export function of(F, M) {
    return (a) => F.of([a, M.empty]);
}
/**
 * @category type class operations
 * @since 3.0.0
 */
export const ap = (F, S) => {
    return ap_(F, W.getApply(S));
};
/**
 * @category type class operations
 * @since 3.0.0
 */
export const chain = (M, S) => (f) => {
    return M.chain(([a, w1]) => pipe(f(a), M.map(([b, w2]) => [b, S.concat(w2)(w1)])));
};
/**
 * @category type class operations
 * @since 3.0.0
 */
export function bimap(F) {
    return (f, g) => F.map(W.bimap(f, g));
}
/**
 * @category type class operations
 * @since 3.0.0
 */
export function mapLeft(F) {
    return (f) => F.map(W.mapLeft(f));
}
// TODO: alt, zero, fromEither, fromReader, fromState, reduce, foldMap, reduceRight, traverse, contramap
// -------------------------------------------------------------------------------------
// utils
// -------------------------------------------------------------------------------------
/**
 * @since 3.0.0
 */
export function fst(F) {
    return F.map(W.fst);
}
/**
 * @since 3.0.0
 */
export function snd(F) {
    return F.map(W.snd);
}
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * @category combinators
 * @since 3.0.0
 */
export function swap(F) {
    return F.map(W.swap);
}
/**
 * Modifies the result to include the changes to the accumulator
 *
 * @category combinators
 * @since 3.0.0
 */
export function listen(F) {
    return F.map(W.listen);
}
/**
 * Applies the returned function to the accumulator
 *
 * @category combinators
 * @since 3.0.0
 */
export function pass(F) {
    return F.map(W.pass);
}
/**
 * Projects a value from modifications made to the accumulator during an action
 *
 * @category combinators
 * @since 3.0.0
 */
export function listens(F) {
    return (f) => F.map(W.listens(f));
}
/**
 * Modify the final accumulator value by applying a function
 *
 * @category combinators
 * @since 3.0.0
 */
export function censor(F) {
    return (f) => F.map(W.censor(f));
}
