/**
 * `Compactable` represents data structures which can be _compacted_/_separated_.
 *
 * @since 3.0.0
 */
import type { Either } from './Either'
import * as FunctorModule from './Functor'
import type { HKT, Kind, Typeclass } from './HKT'
import type { Option } from './Option'
import * as SeparatedModule from './Separated'
import Separated = SeparatedModule.Separated
/**
 * @category type classes
 * @since 3.0.0
 */
export interface Compactable<F extends HKT> extends Typeclass<F> {
  readonly compact: <S, R, W, E, A>(foa: Kind<F, S, R, W, E, Option<A>>) => Kind<F, S, R, W, E, A>
  readonly separate: <S, R, W, E, A, B>(
    fe: Kind<F, S, R, W, E, Either<A, B>>
  ) => Separated<Kind<F, S, R, W, E, A>, Kind<F, S, R, W, E, B>>
}
/**
 * Return a default `compact` implementation from `Functor` and `separate`.
 *
 * @category defaults
 * @since 3.0.0
 */
export declare const compactDefault: <F extends HKT>(
  F: FunctorModule.Functor<F>
) => (
  separate: <S, R, W, E, A, B>(
    fe: Kind<F, S, R, W, E, Either<A, B>>
  ) => SeparatedModule.Separated<Kind<F, S, R, W, E, A>, Kind<F, S, R, W, E, B>>
) => <S_1, R_1, W_1, E_1, A_1>(foa: Kind<F, S_1, R_1, W_1, E_1, Option<A_1>>) => Kind<F, S_1, R_1, W_1, E_1, A_1>
/**
 * Return a default `separate` implementation from `Functor` and `compact`.
 *
 * @category defaults
 * @since 3.0.0
 */
export declare function separateDefault<F extends HKT>(
  F: FunctorModule.Functor<F>
): (compact: Compactable<F>['compact']) => Compactable<F>['separate']
/**
 * `compact` composition.
 *
 * @category combinators
 * @since 3.0.0
 */
export declare function compact<F extends HKT, G extends HKT>(
  F: FunctorModule.Functor<F>,
  G: Compactable<G>
): <FS, FR, FW, FE, GS, GR, GW, GE, A>(
  fgoa: Kind<F, FS, FR, FW, FE, Kind<G, GS, GR, GW, GE, Option<A>>>
) => Kind<F, FS, FR, FW, FE, Kind<G, GS, GR, GW, GE, A>>
/**
 * `separate` composition.
 *
 * @category combinators
 * @since 3.0.0
 */
export declare function separate<F extends HKT, G extends HKT>(
  F: FunctorModule.Functor<F>,
  C: Compactable<G>,
  G: FunctorModule.Functor<G>
): <FS, FR, FW, FE, GS, GR, GW, GE, A, B>(
  fge: Kind<F, FS, FR, FW, FE, Kind<G, GS, GR, GW, GE, Either<A, B>>>
) => Separated<Kind<F, FS, FR, FW, FE, Kind<G, GS, GR, GW, GE, A>>, Kind<F, FS, FR, FW, FE, Kind<G, GS, GR, GW, GE, B>>>
