/**
 * Lift a computation from the `IO` monad.
 *
 * @since 3.0.0
 */
import * as ChainModule from './Chain'
import type { HKT, Kind, Typeclass } from './HKT'
import type { IO } from './IO'
import Chain = ChainModule.Chain
/**
 * @category type classes
 * @since 3.0.0
 */
export interface FromIO<F extends HKT> extends Typeclass<F> {
  readonly fromIO: <A, S, R = unknown, W = never, E = never>(fa: IO<A>) => Kind<F, S, R, W, E, A>
}
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const fromIOK: <F extends HKT>(
  F: FromIO<F>
) => <A extends readonly unknown[], B>(
  f: (...a: A) => IO<B>
) => <S, R = unknown, W = never, E = never>(...a: A) => Kind<F, S, R, W, E, B>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const chainIOK: <M extends HKT>(
  F: FromIO<M>,
  M: ChainModule.Chain<M>
) => <A, B>(f: (a: A) => IO<B>) => <S, R, W, E>(ma: Kind<M, S, R, W, E, A>) => Kind<M, S, R, W, E, B>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const chainFirstIOK: <M extends HKT>(
  F: FromIO<M>,
  M: ChainModule.Chain<M>
) => <A, B>(f: (a: A) => IO<B>) => <S, R, W, E>(first: Kind<M, S, R, W, E, A>) => Kind<M, S, R, W, E, A>
