/**
 * The `FromEither` type class represents those data types which support untyped errors.
 *
 * @since 3.0.0
 */
import type { Chain } from './Chain'
import type { HKT, Kind, Typeclass } from './HKT'
import type { Option } from './Option'
import type { Predicate } from './Predicate'
import type { Refinement } from './Refinement'
/**
 * @category type classes
 * @since 3.0.0
 */
export interface FromOption<F extends HKT> extends Typeclass<F> {
  readonly fromOption: <A, S, R = unknown, W = never, E = never>(fa: Option<A>) => Kind<F, S, R, W, E, A>
}
/**
 * @category constructors
 * @since 3.0.0
 */
export declare const fromPredicate: <F extends HKT>(
  F: FromOption<F>
) => <B extends A, A = B>(
  predicate: Predicate<A>
) => <S, R = unknown, W = never, E = never>(b: B) => Kind<F, S, R, W, E, B>
/**
 * @category constructors
 * @since 3.0.0
 */
export declare const fromRefinement: <F extends HKT>(
  F: FromOption<F>
) => <C extends A, B extends A, A = C>(
  refinement: Refinement<A, B>
) => <S, R = unknown, W = never, E = never>(c: C) => Kind<F, S, R, W, E, B>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const fromOptionK: <F extends HKT>(
  F: FromOption<F>
) => <A extends readonly unknown[], B>(
  f: (...a: A) => Option<B>
) => <S, R = unknown, W = never, E = never>(...a: A) => Kind<F, S, R, W, E, B>
/**
 * @category interop
 * @since 3.0.0
 */
export declare const fromNullable: <F extends HKT>(
  F: FromOption<F>
) => <A, S, R, W, E>(a: A) => Kind<F, S, R, W, E, NonNullable<A>>
/**
 * @category interop
 * @since 3.0.0
 */
export declare const fromNullableK: <F extends HKT>(
  F: FromOption<F>
) => <A extends readonly unknown[], B>(
  f: (...a: A) => B | null | undefined
) => <S, R, W, E>(...a: A) => Kind<F, S, R, W, E, NonNullable<B>>
/**
 * @category interop
 * @since 3.0.0
 */
export declare const chainNullableK: <F extends HKT>(
  F: FromOption<F>,
  C: Chain<F>
) => <A, B>(
  f: (a: A) => B | null | undefined
) => <S, R, W, E>(ma: Kind<F, S, R, W, E, A>) => Kind<F, S, R, W, E, NonNullable<B>>
