/**
 * Lift a computation from the `State` monad.
 *
 * @since 3.0.0
 */
import type { Chain } from './Chain'
import type { Endomorphism } from './Endomorphism'
import type { HKT, Kind, Typeclass } from './HKT'
import * as StateModule from './State'
import State = StateModule.State
/**
 * @category type classes
 * @since 3.0.0
 */
export interface FromState<F extends HKT> extends Typeclass<F> {
  readonly fromState: <S, A, R = unknown, W = never, E = never>(fa: State<S, A>) => Kind<F, S, R, W, E, A>
}
/**
 * @category constructors
 * @since 3.0.0
 */
export declare function get<F extends HKT>(
  F: FromState<F>
): <S, R = unknown, W = never, E = never>() => Kind<F, S, R, W, E, S>
/**
 * @category constructors
 * @since 3.0.0
 */
export declare function put<F extends HKT>(
  F: FromState<F>
): <S, R = unknown, W = never, E = never>(s: S) => Kind<F, S, R, W, E, void>
/**
 * @category constructors
 * @since 3.0.0
 */
export declare const modify: <F extends HKT>(
  F: FromState<F>
) => <S, R = unknown, W = never, E = never>(f: Endomorphism<S>) => Kind<F, S, R, W, E, void>
/**
 * @category constructors
 * @since 3.0.0
 */
export declare const gets: <F extends HKT>(
  F: FromState<F>
) => <S, A, R = unknown, W = never, E = never>(f: (s: S) => A) => Kind<F, S, R, W, E, A>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const fromStateK: <F extends HKT>(
  F: FromState<F>
) => <A extends readonly unknown[], S, B>(
  f: (...a: A) => StateModule.State<S, B>
) => <R = unknown, W = never, E = never>(...a: A) => Kind<F, S, R, W, E, B>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const chainStateK: <M extends HKT>(
  F: FromState<M>,
  M: Chain<M>
) => <A, S, B>(f: (a: A) => StateModule.State<S, B>) => <R, W, E>(ma: Kind<M, S, R, W, E, A>) => Kind<M, S, R, W, E, B>
