import type { HKT, Kind, Typeclass } from './HKT'
/**
 * @category type classes
 * @since 3.0.0
 */
export interface Functor<F extends HKT> extends Typeclass<F> {
  readonly map: <A, B>(f: (a: A) => B) => <S, R, W, E>(fa: Kind<F, S, R, W, E, A>) => Kind<F, S, R, W, E, B>
}
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const flap: <F extends HKT>(
  F: Functor<F>
) => <A>(a: A) => <S, R, W, E, B>(fab: Kind<F, S, R, W, E, (a: A) => B>) => Kind<F, S, R, W, E, B>
/**
 * `map` composition.
 *
 * @category combinators
 * @since 3.0.0
 */
export declare const map: <F extends HKT, G extends HKT>(
  F: Functor<F>,
  G: Functor<G>
) => <A, B>(
  f: (a: A) => B
) => <FS, FR, FW, FE, GS, GR, GW, GE>(
  fga: Kind<F, FS, FR, FW, FE, Kind<G, GS, GR, GW, GE, A>>
) => Kind<F, FS, FR, FW, FE, Kind<G, GS, GR, GW, GE, B>>
/**
 * @since 3.0.0
 */
export declare const bindTo: <F extends HKT>(
  F: Functor<F>
) => <N extends string>(
  name: N
) => <S, R, W, E, A>(fa: Kind<F, S, R, W, E, A>) => Kind<F, S, R, W, E, { readonly [K in N]: A }>
/**
 * @since 3.0.0
 */
export declare const tupled: <F extends HKT>(
  F: Functor<F>
) => <S, R, W, E, A>(fa: Kind<F, S, R, W, E, A>) => Kind<F, S, R, W, E, readonly [A]>
declare const let_: <F extends HKT>(
  F: Functor<F>
) => <N extends string, A, B>(
  name: Exclude<N, keyof A>,
  f: (a: A) => B
) => <S, R, W, E>(
  fa: Kind<F, S, R, W, E, A>
) => Kind<F, S, R, W, E, { readonly [K in N | keyof A]: K extends keyof A ? A[K] : B }>
export {
  /**
   * @since 3.0.0
   */
  let_ as let
}
