/**
 * @since 3.0.0
 */
import * as ApplyModule from './Apply'
import type { Chain } from './Chain'
import type { Either } from './Either'
import { Lazy } from './function'
import * as FunctorModule from './Functor'
import type { HKT, Kind } from './HKT'
import type { Monad } from './Monad'
import * as OptionModule from './Option'
import type { Pointed } from './Pointed'
import Option = OptionModule.Option
import Functor = FunctorModule.Functor
/**
 * @category constructors
 * @since 3.0.0
 */
export declare const some: <F extends HKT>(
  F: Pointed<F>
) => <A, S, R, W, E>(a: A) => Kind<F, S, R, W, E, OptionModule.Option<A>>
/**
 * @category natural transformations
 * @since 3.0.0
 */
export declare function fromF<F extends HKT>(
  F: Functor<F>
): <S, R, W, E, A>(ma: Kind<F, S, R, W, E, A>) => Kind<F, S, R, W, E, Option<A>>
/**
 * @category natural transformations
 * @since 3.0.0
 */
export declare const fromEither: <F extends HKT>(
  F: Pointed<F>
) => <A, S, R, W, E>(e: Either<unknown, A>) => Kind<F, S, R, W, E, OptionModule.Option<A>>
/**
 * @since 3.0.0
 */
export declare function match<F extends HKT>(
  F: Functor<F>
): <B, A, C = B>(
  onNone: () => B,
  onSome: (a: A) => C
) => <S, R, W, E>(ma: Kind<F, S, R, W, E, Option<A>>) => Kind<F, S, R, W, E, B | C>
/**
 * @since 3.0.0
 */
export declare const matchE: <M extends HKT>(
  M: Chain<M>
) => <S, R2, W2, E2, B, A, R3, W3, E3, C = B>(
  onNone: () => Kind<M, S, R2, W2, E2, B>,
  onSome: (a: A) => Kind<M, S, R3, W3, E3, C>
) => <R1, W1, E1>(
  ma: Kind<M, S, R1, W1, E1, OptionModule.Option<A>>
) => Kind<M, S, R1 & R2 & R3, W2 | W3 | W1, E2 | E3 | E1, B | C>
/**
 * @since 3.0.0
 */
export declare const getOrElse: <F extends HKT>(
  F: FunctorModule.Functor<F>
) => <B>(
  onNone: Lazy<B>
) => <S, R, W, E, A>(fa: Kind<F, S, R, W, E, OptionModule.Option<A>>) => Kind<F, S, R, W, E, B | A>
/**
 * @since 3.0.0
 */
export declare const getOrElseE: <M extends HKT>(
  M: Monad<M>
) => <S, R2, W2, E2, B>(
  onNone: Lazy<Kind<M, S, R2, W2, E2, B>>
) => <R1, W1, E1, A>(ma: Kind<M, S, R1, W1, E1, OptionModule.Option<A>>) => Kind<M, S, R1 & R2, W2 | W1, E2 | E1, B | A>
/**
 * @since 3.0.0
 */
export declare function map<F extends HKT>(
  F: Functor<F>
): <A, B>(f: (a: A) => B) => <S, R, W, E>(fa: Kind<F, S, R, W, E, Option<A>>) => Kind<F, S, R, W, E, Option<B>>
/**
 * @since 3.0.0
 */
export declare const ap: <F extends HKT>(
  F: ApplyModule.Apply<F>
) => <S, R2, W2, E2, A>(
  fa: Kind<F, S, R2, W2, E2, OptionModule.Option<A>>
) => <R1, W1, E1, B>(
  fab: Kind<F, S, R1, W1, E1, OptionModule.Option<(a: A) => B>>
) => Kind<F, S, R1 & R2, W2 | W1, E2 | E1, OptionModule.Option<B>>
/**
 * @since 3.0.0
 */
export declare const chain: <M extends HKT>(
  M: Monad<M>
) => <A, S, R, W, E, B>(
  f: (a: A) => Kind<M, S, R, W, E, OptionModule.Option<B>>
) => (ma: Kind<M, S, R, W, E, OptionModule.Option<A>>) => Kind<M, S, R, W, E, OptionModule.Option<B>>
/**
 * @since 3.0.0
 */
export declare const alt: <M extends HKT>(
  M: Monad<M>
) => <S, R2, W2, E2, B>(
  second: Lazy<Kind<M, S, R2, W2, E2, OptionModule.Option<B>>>
) => <R1, W1, E1, A>(
  first: Kind<M, S, R1, W1, E1, OptionModule.Option<A>>
) => Kind<M, S, R1 & R2, W2 | W1, E2 | E1, OptionModule.Option<B | A>>
/**
 * @since 3.0.0
 */
export declare function zero<F extends HKT>(F: Pointed<F>): <S, R, W, E, A>() => Kind<F, S, R, W, E, Option<A>>
