/**
 * @since 3.0.0
 */
import * as AltModule from './Alt'
import * as ApplicativeModule from './Applicative'
import * as ApplyModule from './Apply'
import * as BifunctorModule from './Bifunctor'
import * as ChainModule from './Chain'
import * as CompactableModule from './Compactable'
import * as EitherModule from './Either'
import * as FilterableModule from './Filterable'
import * as FromEitherModule from './FromEither'
import * as FromReaderModule from './FromReader'
import { Lazy } from './function'
import * as FunctorModule from './Functor'
import type { HKT } from './HKT'
import * as MonadModule from './Monad'
import type { Monoid } from './Monoid'
import type { Option } from './Option'
import * as PointedModule from './Pointed'
import type { Predicate } from './Predicate'
import * as ReaderModule from './Reader'
import type { ReadonlyNonEmptyArray } from './ReadonlyNonEmptyArray'
import type { Refinement } from './Refinement'
import type { Semigroup } from './Semigroup'
import Either = EitherModule.Either
import Reader = ReaderModule.Reader
/**
 * @category model
 * @since 3.0.0
 */
export interface ReaderEither<R, E, A> extends Reader<R, Either<E, A>> {}
/**
 * @category constructors
 * @since 3.0.0
 */
export declare const left: <E, R = unknown, A = never>(e: E) => ReaderEither<R, E, A>
/**
 * @category constructors
 * @since 3.0.0
 */
export declare const right: <A, R = unknown, E = never>(a: A) => ReaderEither<R, E, A>
/**
 * @category constructors
 * @since 3.0.0
 */
export declare const rightReader: <R, A, E = never>(ma: Reader<R, A>) => ReaderEither<R, E, A>
/**
 * @category constructors
 * @since 3.0.0
 */
export declare const leftReader: <R, E, A = never>(me: Reader<R, E>) => ReaderEither<R, E, A>
/**
 * @category constructors
 * @since 3.0.0
 */
export declare const asksReaderEither: <R1, R2, E, A>(
  f: (r1: R1) => ReaderEither<R2, E, A>
) => ReaderEither<R1 & R2, E, A>
/**
 * @category natural transformations
 * @since 3.0.0
 */
export declare const fromEither: <E, A, R = unknown>(fa: Either<E, A>) => ReaderEither<R, E, A>
/**
 * @category natural transformations
 * @since 3.0.0
 */
export declare const fromReader: <R, A, E = never>(fa: Reader<R, A>) => ReaderEither<R, E, A>
/**
 * @category destructors
 * @since 3.0.0
 */
export declare const match: <E, B, A, C = B>(
  onLeft: (e: E) => B,
  onRight: (a: A) => C
) => <R>(ma: Reader<R, EitherModule.Either<E, A>>) => Reader<R, B | C>
/**
 * @category destructors
 * @since 3.0.0
 */
export declare const matchE: <E, R2, B, A, R3, C = B>(
  onLeft: (e: E) => Reader<R2, B>,
  onRight: (a: A) => Reader<R3, C>
) => <R1>(ma: Reader<R1, EitherModule.Either<E, A>>) => Reader<R1 & R2 & R3, B | C>
/**
 * @category destructors
 * @since 3.0.0
 */
export declare const getOrElse: <E, B>(onLeft: (e: E) => B) => <R, A>(ma: ReaderEither<R, E, A>) => Reader<R, A | B>
/**
 * @category destructors
 * @since 3.0.0
 */
export declare const getOrElseE: <E, R2, B>(
  onLeft: (e: E) => Reader<R2, B>
) => <R1, A>(ma: ReaderEither<R1, E, A>) => Reader<R1 & R2, A | B>
/**
 * @category interop
 * @since 3.0.0
 */
export declare const toUnion: <R, E, A>(fa: ReaderEither<R, E, A>) => Reader<R, E | A>
/**
 * Changes the value of the local context during the execution of the action `ma` (similar to `Contravariant`'s
 * `contramap`).
 *
 * @category combinators
 * @since 3.0.0
 */
export declare const local: <R2, R1>(f: (r2: R2) => R1) => <E, A>(ma: ReaderEither<R1, E, A>) => ReaderEither<R2, E, A>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const orElse: <E1, R1, E2, B>(
  onLeft: (e: E1) => ReaderEither<R1, E2, B>
) => <R2, A>(ma: ReaderEither<R2, E1, A>) => ReaderEither<R1 & R2, E2, A | B>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const orElseFirst: <E1, R2, E2, B>(
  onLeft: (e: E1) => ReaderEither<R2, E2, B>
) => <R1, A>(ma: ReaderEither<R1, E1, A>) => ReaderEither<R1 & R2, E1 | E2, A>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const orLeft: <E1, R, E2>(
  onLeft: (e: E1) => Reader<R, E2>
) => <A>(fa: ReaderEither<R, E1, A>) => ReaderEither<R, E2, A>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const swap: <R, E, A>(ma: ReaderEither<R, E, A>) => ReaderEither<R, A, E>
/**
 * `map` can be used to turn functions `(a: A) => B` into functions `(fa: F<A>) => F<B>` whose argument and return types
 * use the type constructor `F` to represent some computational context.
 *
 * @category Functor
 * @since 3.0.0
 */
export declare const map: <A, B>(f: (a: A) => B) => <R, E>(fa: ReaderEither<R, E, A>) => ReaderEither<R, E, B>
/**
 * Map a pair of functions over the two last type arguments of the bifunctor.
 *
 * @category Bifunctor
 * @since 3.0.0
 */
export declare const bimap: <E, G, A, B>(
  f: (e: E) => G,
  g: (a: A) => B
) => <R>(fea: ReaderEither<R, E, A>) => ReaderEither<R, G, B>
/**
 * Map a function over the second type argument of a bifunctor.
 *
 * @category Bifunctor
 * @since 3.0.0
 */
export declare const mapLeft: <E, G>(f: (e: E) => G) => <R, A>(fea: ReaderEither<R, E, A>) => ReaderEither<R, G, A>
/**
 * Apply a function to an argument under a type constructor.
 *
 * @category Apply
 * @since 3.0.0
 */
export declare const ap: <R2, E2, A>(
  fa: ReaderEither<R2, E2, A>
) => <R1, E1, B>(fab: ReaderEither<R1, E1, (a: A) => B>) => ReaderEither<R1 & R2, E1 | E2, B>
/**
 * @category Pointed
 * @since 3.0.0
 */
export declare const of: <A, R = unknown, E = never>(a: A) => ReaderEither<R, E, A>
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation.
 *
 * @category Chain
 * @since 3.0.0
 */
export declare const chain: <A, R2, E2, B>(
  f: (a: A) => ReaderEither<R2, E2, B>
) => <R1, E1>(ma: ReaderEither<R1, E1, A>) => ReaderEither<R1 & R2, E1 | E2, B>
/**
 * Derivable from `Chain`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export declare const flatten: <R1, E1, R2, E2, A>(
  mma: ReaderEither<R1, E1, ReaderEither<R2, E2, A>>
) => ReaderEither<R1 & R2, E1 | E2, A>
/**
 * Identifies an associative operation on a type constructor. It is similar to `Semigroup`, except that it applies to
 * types of kind `* -> *`.
 *
 * @category Alt
 * @since 3.0.0
 */
export declare const alt: <R2, E2, B>(
  second: () => ReaderEither<R2, E2, B>
) => <R1, E1, A>(first: ReaderEither<R1, E1, A>) => ReaderEither<R1 & R2, E2, A | B>
/**
 * @category HKT
 * @since 3.0.0
 */
export interface ReaderEitherF extends HKT {
  readonly type: ReaderEither<this['Contravariant1'], this['Covariant2'], this['Covariant1']>
}
/**
 * @category HKT
 * @since 3.0.0
 */
export interface ReaderEitherFFixedE<E> extends HKT {
  readonly type: ReaderEither<this['Contravariant1'], E, this['Covariant1']>
}
/**
 * The default [`Applicative`](#applicative) instance returns the first error, if you want to
 * get all errors you need to provide an way to concatenate them via a `Semigroup`.
 *
 * See [`getApplicativeValidation`](./Either.ts.html#getapplicativevalidation).
 *
 * @category instances
 * @since 3.0.0
 */
export declare const getApplicativeReaderValidation: <E>(
  S: Semigroup<E>
) => ApplicativeModule.Applicative<ReaderEitherFFixedE<E>>
/**
 * The default [`Alt`](#alt) instance returns the last error, if you want to
 * get all errors you need to provide an way to concatenate them via a `Semigroup`.
 *
 * See [`getAltValidation`](./Either.ts.html#getaltvalidation).
 *
 * @category instances
 * @since 3.0.0
 */
export declare const getAltReaderValidation: <E>(S: Semigroup<E>) => AltModule.Alt<ReaderEitherFFixedE<E>>
/**
 * @category instances
 * @since 3.0.0
 */
export declare const getCompactable: <E>(M: Monoid<E>) => CompactableModule.Compactable<ReaderEitherFFixedE<E>>
/**
 * @category instances
 * @since 3.0.0
 */
export declare const getFilterable: <E>(M: Monoid<E>) => FilterableModule.Filterable<ReaderEitherFFixedE<E>>
/**
 * @category instances
 * @since 3.0.0
 */
export declare const Functor: FunctorModule.Functor<ReaderEitherF>
/**
 * Derivable from `Functor`.
 *
 * @category combinators
 * @since 3.0.0
 */
export declare const flap: <A>(a: A) => <R, E, B>(fab: ReaderEither<R, E, (a: A) => B>) => ReaderEither<R, E, B>
/**
 * @category instances
 * @since 3.0.0
 */
export declare const Pointed: PointedModule.Pointed<ReaderEitherF>
/**
 * @category instances
 * @since 3.0.0
 */
export declare const Apply: ApplyModule.Apply<ReaderEitherF>
/**
 * Combine two effectful actions, keeping only the result of the first.
 *
 * Derivable from `Apply`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export declare const apFirst: <R2, E2, B>(
  second: ReaderEither<R2, E2, B>
) => <R1, E1, A>(first: ReaderEither<R1, E1, A>) => ReaderEither<R1 & R2, E1 | E2, A>
/**
 * Combine two effectful actions, keeping only the result of the second.
 *
 * Derivable from `Apply`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export declare const apSecond: <R2, E2, B>(
  second: ReaderEither<R2, E2, B>
) => <R1, E1, A>(first: ReaderEither<R1, E1, A>) => ReaderEither<R1 & R2, E1 | E2, B>
/**
 * @category instances
 * @since 3.0.0
 */
export declare const Applicative: ApplicativeModule.Applicative<ReaderEitherF>
/**
 * @category instances
 * @since 3.0.0
 */
export declare const Chain: ChainModule.Chain<ReaderEitherF>
/**
 * @category instances
 * @since 3.0.0
 */
export declare const Monad: MonadModule.Monad<ReaderEitherF>
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation and
 * keeping only the result of the first.
 *
 * Derivable from `Chain`.
 *
 * @category derivable combinators
 * @since 3.0.0
 */
export declare const chainFirst: <A, R2, E2, B>(
  f: (a: A) => ReaderEither<R2, E2, B>
) => <R1, E1>(first: ReaderEither<R1, E1, A>) => ReaderEither<R1 & R2, E1 | E2, A>
/**
 * @category instances
 * @since 3.0.0
 */
export declare const Bifunctor: BifunctorModule.Bifunctor<ReaderEitherF>
/**
 * @category instances
 * @since 3.0.0
 */
export declare const Alt: AltModule.Alt<ReaderEitherF>
/**
 * @category instances
 * @since 3.0.0
 */
export declare const FromReader: FromReaderModule.FromReader<ReaderEitherF>
/**
 * Reads the current context.
 *
 * @category constructors
 * @since 3.0.0
 */
export declare const ask: <R, E = never>() => ReaderEither<R, E, R>
/**
 * Projects a value from the global context in a `ReaderEither`.
 *
 * @category constructors
 * @since 3.0.0
 */
export declare const asks: <R, A, E = never>(f: (r: R) => A) => ReaderEither<R, E, A>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const fromReaderK: <A extends ReadonlyArray<unknown>, R, B>(
  f: (...a: A) => Reader<R, B>
) => <E = never>(...a: A) => ReaderEither<R, E, B>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const chainReaderK: <A, R2, B>(
  f: (a: A) => Reader<R2, B>
) => <R1, E = never>(ma: ReaderEither<R1, E, A>) => ReaderEither<R1 & R2, E, B>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const chainFirstReaderK: <A, R2, B>(
  f: (a: A) => Reader<R2, B>
) => <R1, E = never>(ma: ReaderEither<R1, E, A>) => ReaderEither<R1 & R2, E, A>
/**
 * @category instances
 * @since 3.0.0
 */
export declare const FromEither: FromEitherModule.FromEither<ReaderEitherF>
/**
 * Derivable from `FromEither`.
 *
 * @category natural transformations
 * @since 3.0.0
 */
export declare const fromOption: <E>(onNone: Lazy<E>) => <A, R = unknown>(fa: Option<A>) => ReaderEither<R, E, A>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const fromOptionKOrElse: <E>(
  onNone: Lazy<E>
) => <A extends ReadonlyArray<unknown>, B>(f: (...a: A) => Option<B>) => <R = unknown>(...a: A) => ReaderEither<R, E, B>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const chainOptionKOrElse: <E>(
  onNone: Lazy<E>
) => <A, B>(f: (a: A) => Option<B>) => <R>(ma: ReaderEither<R, E, A>) => ReaderEither<R, E, B>
/**
 * Derivable from `FromEither`.
 *
 * @category constructors
 * @since 3.0.0
 */
export declare const fromPredicateOrElse: <B extends A, E, A = B>(
  predicate: Predicate<A>,
  onFalse: (b: B) => E
) => <R = unknown>(b: B) => ReaderEither<R, E, B>
/**
 * @category constructors
 * @since 3.0.0
 */
export declare const fromRefinementOrElse: <C extends A, B extends A, E, A = C>(
  refinement: Refinement<A, B>,
  onFalse: (c: C) => E
) => <R = unknown>(c: C) => ReaderEither<R, E, B>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const filterOrElse: <B extends A, E2, A = B>(
  predicate: Predicate<A>,
  onFalse: (b: B) => E2
) => <R, E1>(mb: ReaderEither<R, E1, B>) => ReaderEither<R, E2 | E1, B>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const refineOrElse: <C extends A, B extends A, E2, A = C>(
  refinement: Refinement<A, B>,
  onFalse: (c: C) => E2
) => <R, E1>(ma: ReaderEither<R, E1, C>) => ReaderEither<R, E2 | E1, B>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const fromEitherK: <A extends ReadonlyArray<unknown>, E, B>(
  f: (...a: A) => EitherModule.Either<E, B>
) => <R = unknown>(...a: A) => ReaderEither<R, E, B>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const chainEitherK: <A, E2, B>(
  f: (a: A) => Either<E2, B>
) => <R, E1>(ma: ReaderEither<R, E1, A>) => ReaderEither<R, E1 | E2, B>
/**
 * @category combinators
 * @since 3.0.0
 */
export declare const chainFirstEitherK: <A, E2, B>(
  f: (a: A) => Either<E2, B>
) => <R, E1>(ma: ReaderEither<R, E1, A>) => ReaderEither<R, E1 | E2, A>
/**
 * @category interop
 * @since 3.0.0
 */
export declare const fromNullableOrElse: <E>(
  onNullable: Lazy<E>
) => <A, R = unknown>(a: A) => ReaderEither<R, E, NonNullable<A>>
/**
 * @category interop
 * @since 3.0.0
 */
export declare const fromNullableKOrElse: <E>(
  onNullable: Lazy<E>
) => <A extends ReadonlyArray<unknown>, B>(
  f: (...a: A) => B | null | undefined
) => <R = unknown>(...a: A) => ReaderEither<R, E, NonNullable<B>>
/**
 * @category interop
 * @since 3.0.0
 */
export declare const chainNullableKOrElse: <E>(
  onNullable: Lazy<E>
) => <A, B>(f: (a: A) => B | null | undefined) => <R>(ma: ReaderEither<R, E, A>) => ReaderEither<R, E, NonNullable<B>>
/**
 * @since 3.0.0
 */
export declare const Do: ReaderEither<unknown, never, {}>
/**
 * @since 3.0.0
 */
export declare const bindTo: <N extends string>(
  name: N
) => <R, E, A>(
  fa: ReaderEither<R, E, A>
) => ReaderEither<
  R,
  E,
  {
    readonly [K in N]: A
  }
>
declare const let_: <N extends string, A, B>(
  name: Exclude<N, keyof A>,
  f: (a: A) => B
) => <R, E>(
  fa: ReaderEither<R, E, A>
) => ReaderEither<
  R,
  E,
  {
    readonly [K in N | keyof A]: K extends keyof A ? A[K] : B
  }
>
export {
  /**
   * @since 3.0.0
   */
  let_ as let
}
/**
 * @since 3.0.0
 */
export declare const bind: <N extends string, A, R2, E2, B>(
  name: Exclude<N, keyof A>,
  f: <A2 extends A>(a: A | A2) => ReaderEither<R2, E2, B>
) => <R1, E1>(
  fa: ReaderEither<R1, E1, A>
) => ReaderEither<
  R1 & R2,
  E1 | E2,
  {
    readonly [K in keyof A | N]: K extends keyof A ? A[K] : B
  }
>
/**
 * @since 3.0.0
 */
export declare const apS: <N extends string, A, R2, E2, B>(
  name: Exclude<N, keyof A>,
  fb: ReaderEither<R2, E2, B>
) => <R1, E1>(
  fa: ReaderEither<R1, E1, A>
) => ReaderEither<
  R1 & R2,
  E1 | E2,
  {
    readonly [K in keyof A | N]: K extends keyof A ? A[K] : B
  }
>
/**
 * @since 3.0.0
 */
export declare const ApT: ReaderEither<unknown, never, readonly []>
/**
 * @since 3.0.0
 */
export declare const tupled: <R, E, A>(fa: ReaderEither<R, E, A>) => ReaderEither<R, E, readonly [A]>
/**
 * @since 3.0.0
 */
export declare const apT: <R2, E2, B>(
  fb: ReaderEither<R2, E2, B>
) => <R1, E1, A extends ReadonlyArray<unknown>>(
  fas: ReaderEither<R1, E1, A>
) => ReaderEither<R1 & R2, E1 | E2, readonly [...A, B]>
/**
 * Make sure that a resource is cleaned up in the event of an exception (\*). The release action is called regardless of
 * whether the body action throws (\*) or returns.
 *
 * (\*) i.e. returns a `Left`
 *
 * @since 3.0.0
 */
export declare const bracket: <R, E, A, B>(
  aquire: ReaderEither<R, E, A>,
  use: (a: A) => ReaderEither<R, E, B>,
  release: (a: A, e: Either<E, B>) => ReaderEither<R, E, void>
) => ReaderEither<R, E, B>
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverseWithIndex(Apply)`.
 *
 * @since 3.0.0
 */
export declare const traverseReadonlyNonEmptyArrayWithIndex: <A, R, E, B>(
  f: (index: number, a: A) => ReaderEither<R, E, B>
) => (as: ReadonlyNonEmptyArray<A>) => ReaderEither<R, E, ReadonlyNonEmptyArray<B>>
/**
 * Equivalent to `ReadonlyArray#traverseWithIndex(Applicative)`.
 *
 * @since 3.0.0
 */
export declare const traverseReadonlyArrayWithIndex: <A, R, E, B>(
  f: (index: number, a: A) => ReaderEither<R, E, B>
) => (as: readonly A[]) => ReaderEither<R, E, readonly B[]>
/**
 * Equivalent to `ReadonlyNonEmptyArray#traverse(Apply)`.
 *
 * @since 3.0.0
 */
export declare const traverseReadonlyNonEmptyArray: <A, R, E, B>(
  f: (a: A) => ReaderEither<R, E, B>
) => (as: ReadonlyNonEmptyArray<A>) => ReaderEither<R, E, ReadonlyNonEmptyArray<B>>
/**
 * Equivalent to `ReadonlyArray#traverse(Applicative)`.
 *
 * @since 3.0.0
 */
export declare const traverseReadonlyArray: <A, R, E, B>(
  f: (a: A) => ReaderEither<R, E, B>
) => (as: readonly A[]) => ReaderEither<R, E, readonly B[]>
/**
 * Equivalent to `ReadonlyArray#sequence(Applicative)`.
 *
 * @since 3.0.0
 */
export declare const sequenceReadonlyArray: <R, E, A>(
  arr: ReadonlyArray<ReaderEither<R, E, A>>
) => ReaderEither<R, E, ReadonlyArray<A>>
