/**
 * @since 3.0.0
 */
import type { Apply } from './Apply'
import type { Chain } from './Chain'
import type { Functor } from './Functor'
import type { HKT, Kind } from './HKT'
import type { NaturalTransformation } from './NaturalTransformation'
import type { Pointed } from './Pointed'
import type { Reader } from './Reader'
/**
 * @since 3.0.0
 */
export declare function of<F extends HKT>(
  F: Pointed<F>
): <A, R, S, FR, W, E>(a: A) => Reader<R, Kind<F, S, FR, W, E, A>>
/**
 * @since 3.0.0
 */
export declare function map<F extends HKT>(
  F: Functor<F>
): <A, B>(
  f: (a: A) => B
) => <R, S, FR, W, E>(fa: Reader<R, Kind<F, S, FR, W, E, A>>) => Reader<R, Kind<F, S, FR, W, E, B>>
/**
 * @since 3.0.0
 */
export declare const ap: <F extends HKT>(
  F: Apply<F>
) => <R2, S, FR2, W2, E2, A>(
  fa: Reader<R2, Kind<F, S, FR2, W2, E2, A>>
) => <R1, FR1, W1, E1, B>(
  fab: Reader<R1, Kind<F, S, FR1, W1, E1, (a: A) => B>>
) => Reader<R1 & R2, Kind<F, S, FR1 & FR2, W2 | W1, E2 | E1, B>>
/**
 * @since 3.0.0
 */
export declare const chain: <M extends HKT>(
  M: Chain<M>
) => <A, R2, S, FR2, W2, E2, B>(
  f: (a: A) => Reader<R2, Kind<M, S, FR2, W2, E2, B>>
) => <R1, FR1, W1, E1>(
  ma: Reader<R1, Kind<M, S, FR1, W1, E1, A>>
) => Reader<R1 & R2, Kind<M, S, FR1 & FR2, W2 | W1, E2 | E1, B>>
/**
 * @category constructors
 * @since 3.0.0
 */
export declare const fromReader: <F extends HKT>(
  F: Pointed<F>
) => <R, A, S, FR = unknown, W = never, E = never>(ma: Reader<R, A>) => Reader<R, Kind<F, S, FR, W, E, A>>
/**
 * @category constructors
 * @since 3.0.0
 */
export declare function fromNaturalTransformation<F extends HKT, G extends HKT>(
  nt: NaturalTransformation<F, G>
): <R, S, FR, W, E, A>(f: (r: R) => Kind<F, S, FR, W, E, A>) => Reader<R, Kind<G, S, FR, W, E, A>>
