/**
 * @since 3.0.0
 */
import type { Chain } from './Chain'
import type { Functor } from './Functor'
import type { HKT, Kind } from './HKT'
import type { Pointed } from './Pointed'
import type { State } from './State'
/**
 * @since 3.0.0
 */
export interface StateT<F extends HKT, FS, FR, FW, FE, S, A> {
  (s: S): Kind<F, FS, FR, FW, FE, readonly [A, S]>
}
/**
 * @since 3.0.0
 */
export declare function of<F extends HKT>(
  F: Pointed<F>
): <A, FS, FR, FW, FE, S>(a: A) => StateT<F, FS, FR, FW, FE, S, A>
/**
 * @since 3.0.0
 */
export declare function map<F extends HKT>(
  F: Functor<F>
): <A, B>(f: (a: A) => B) => <FS, FR, FW, FE, S>(fa: StateT<F, FS, FR, FW, FE, S, A>) => StateT<F, FS, FR, FW, FE, S, B>
/**
 * @since 3.0.0
 */
export declare const ap: <F extends HKT>(
  F: Chain<F>
) => <FS, FR2, FW2, FE2, S, A>(
  fa: StateT<F, FS, FR2, FW2, FE2, S, A>
) => <FR1, FW1, FE1, B>(
  fab: StateT<F, FS, FR1, FW1, FE1, S, (a: A) => B>
) => StateT<F, FS, FR1 & FR2, FW2 | FW1, FE2 | FE1, S, B>
/**
 * @since 3.0.0
 */
export declare const chain: <F extends HKT>(
  F: Chain<F>
) => <A, FS, FR2, FW2, FE2, S, B>(
  f: (a: A) => StateT<F, FS, FR2, FW2, FE2, S, B>
) => <FR1, FW1, FE1>(ma: StateT<F, FS, FR1, FW1, FE1, S, A>) => StateT<F, FS, FR1 & FR2, FW2 | FW1, FE2 | FE1, S, B>
/**
 * @since 3.0.0
 */
export declare function fromState<F extends HKT>(
  F: Pointed<F>
): <S, A, FS, FR, FW, FE>(sa: State<S, A>) => StateT<F, FS, FR, FW, FE, S, A>
/**
 * @since 3.0.0
 */
export declare function fromF<F extends HKT>(
  F: Functor<F>
): <FS, FR, FW, FE, A, S>(ma: Kind<F, FS, FR, FW, FE, A>) => StateT<F, FS, FR, FW, FE, S, A>
/**
 * @since 3.0.0
 */
export declare function evaluate<F extends HKT>(
  F: Functor<F>
): <S>(s: S) => <FS, FR, FW, FE, A>(ma: StateT<F, FS, FR, FW, FE, S, A>) => Kind<F, FS, FR, FW, FE, A>
/**
 * @since 3.0.0
 */
export declare function execute<F extends HKT>(
  F: Functor<F>
): <S>(s: S) => <FS, FR, FW, FE, A>(ma: StateT<F, FS, FR, FW, FE, S, A>) => Kind<F, FS, FR, FW, FE, S>
