/**
 * @since 3.0.0
 */
import { Apply } from './Apply'
import type { Chain } from './Chain'
import { Lazy } from './function'
import { Functor } from './Functor'
import type { HKT, Kind } from './HKT'
import type { Monad } from './Monad'
import type { Pointed } from './Pointed'
import type { Semigroup } from './Semigroup'
import * as T from './These'
import These = T.These
/**
 * @since 3.0.0
 */
export declare const right: <F extends HKT>(
  F: Pointed<F>
) => <A, S, R = unknown, W = never, FE = never, E = never>(a: A) => Kind<F, S, R, W, FE, T.These<E, A>>
/**
 * @since 3.0.0
 */
export declare const left: <F extends HKT>(
  F: Pointed<F>
) => <E, S, R = unknown, W = never, FE = never, A = never>(e: E) => Kind<F, S, R, W, FE, T.These<E, A>>
/**
 * @since 3.0.0
 */
export declare const both: <F extends HKT>(
  F: Pointed<F>
) => <E, A, S, R = unknown, W = never, FE = never>(e: E, a: A) => Kind<F, S, R, W, FE, T.These<E, A>>
/**
 * @since 3.0.0
 */
export declare function rightF<F extends HKT>(
  F: Functor<F>
): <S, R, W, FE, A, E = never>(fa: Kind<F, S, R, W, FE, A>) => Kind<F, S, R, W, FE, These<E, A>>
/**
 * @since 3.0.0
 */
export declare function leftF<F extends HKT>(
  F: Functor<F>
): <S, R, W, FE, E, A = never>(fe: Kind<F, S, R, W, FE, E>) => Kind<F, S, R, W, FE, These<E, A>>
/**
 * @since 3.0.0
 */
export declare function map<F extends HKT>(
  F: Functor<F>
): <A, B>(
  f: (a: A) => B
) => <S, R, W, FE, E>(fa: Kind<F, S, R, W, FE, These<E, A>>) => Kind<F, S, R, W, FE, These<E, B>>
/**
 * @since 3.0.0
 */
export declare const ap: <F extends HKT, E>(
  F: Apply<F>,
  S: Semigroup<E>
) => <S, R2, W2, FE2, A>(
  fa: Kind<F, S, R2, W2, FE2, T.These<E, A>>
) => <R1, W1, FE1, B>(
  fab: Kind<F, S, R1, W1, FE1, T.These<E, (a: A) => B>>
) => Kind<F, S, R1 & R2, W2 | W1, FE2 | FE1, T.These<E, B>>
/**
 * @since 3.0.0
 */
export declare const chain: <M extends HKT, E>(
  M: Monad<M>,
  S: Semigroup<E>
) => <A, S, R2, W2, FE2, B>(
  f: (a: A) => Kind<M, S, R2, W2, FE2, T.These<E, B>>
) => <R1, W1, FE1>(ma: Kind<M, S, R1, W1, FE1, T.These<E, A>>) => Kind<M, S, R1 & R2, W2 | W1, FE2 | FE1, T.These<E, B>>
/**
 * @since 3.0.0
 */
export declare function bimap<F extends HKT>(
  F: Functor<F>
): <E, G, A, B>(
  f: (e: E) => G,
  g: (a: A) => B
) => <S, R, W, FE>(fea: Kind<F, S, R, W, FE, These<E, A>>) => Kind<F, S, R, W, FE, These<G, B>>
/**
 * @since 3.0.0
 */
export declare function mapLeft<F extends HKT>(
  F: Functor<F>
): <E, G>(
  f: (e: E) => G
) => <S, R, W, FE, A>(fea: Kind<F, S, R, W, FE, These<E, A>>) => Kind<F, S, R, W, FE, These<G, A>>
/**
 * @since 3.0.0
 */
export declare function match<F extends HKT>(
  F: Functor<F>
): <E, B, A, C = B, D = B>(
  onLeft: (e: E) => B,
  onRight: (a: A) => C,
  onBoth: (e: E, a: A) => D
) => <S, R, W, FE>(ma: Kind<F, S, R, W, FE, These<E, A>>) => Kind<F, S, R, W, FE, B | C | D>
/**
 * @since 3.0.0
 */
export declare const matchE: <M extends HKT>(
  M: Chain<M>
) => <E, S, R2, W2, FE2, B, A, R3, W3, FE3, R4, W4, FE4, C = B, D = B>(
  onLeft: (e: E) => Kind<M, S, R2, W2, FE2, B>,
  onRight: (a: A) => Kind<M, S, R3, W3, FE3, C>,
  onBoth: (e: E, a: A) => Kind<M, S, R4, W4, FE4, D>
) => <R1, W1, FE1>(
  ma: Kind<M, S, R1, W1, FE1, T.These<E, A>>
) => Kind<M, S, R1 & R2 & R3 & R4, W2 | W3 | W4 | W1, FE2 | FE3 | FE4 | FE1, B | C | D>
/**
 * @since 3.0.0
 */
export declare function swap<F extends HKT>(
  F: Functor<F>
): <S, R, W, FE, E, A>(ma: Kind<F, S, R, W, FE, These<E, A>>) => Kind<F, S, R, W, FE, These<A, E>>
/**
 * @since 3.0.0
 */
export declare function toTuple2<F extends HKT>(
  F: Functor<F>
): <E, A>(
  e: Lazy<E>,
  a: Lazy<A>
) => <S, R, W, FE>(fa: Kind<F, S, R, W, FE, These<E, A>>) => Kind<F, S, R, W, FE, readonly [E, A]>
